# Copyright 2012 Pinterest.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import pytest

try:
    import pylibmc  # type: ignore

    HAS_PYLIBMC = True
except Exception:
    HAS_PYLIBMC = False

try:
    import memcache  # type: ignore

    HAS_MEMCACHE = True
except Exception:
    HAS_MEMCACHE = False


try:
    import pymemcache.client

    HAS_PYMEMCACHE = True
except Exception:
    HAS_PYMEMCACHE = False


@pytest.fixture(
    params=[
        "pylibmc",
        "memcache",
        "pymemcache",
    ]
)
def client(request, host, port):
    if request.param == "pylibmc":
        if not HAS_PYLIBMC:
            pytest.skip("requires pylibmc")
        client = pylibmc.Client([f"{host}:{port}"])
        client.behaviors = {"tcp_nodelay": True}

    elif request.param == "memcache":
        if not HAS_MEMCACHE:
            pytest.skip("requires python-memcached")
        client = memcache.Client([f"{host}:{port}"])

    elif request.param == "pymemcache":
        if not HAS_PYMEMCACHE:
            pytest.skip("requires pymemcache")
        client = pymemcache.client.Client((host, port))

    else:
        pytest.skip(f"unknown library {request.param}")

    client.flush_all()
    return client


def benchmark(count, func, *args, **kwargs):
    start = time.time()

    for _ in range(count):
        result = func(*args, **kwargs)

    duration = time.time() - start
    print(str(duration))

    return result


@pytest.mark.benchmark()
def test_bench_get(request, client, pairs, count):
    key = "pymemcache_test:0"
    value = pairs[key]
    client.set(key, value)
    benchmark(count, client.get, key)


@pytest.mark.benchmark()
def test_bench_set(request, client, pairs, count):
    key = "pymemcache_test:0"
    value = pairs[key]
    benchmark(count, client.set, key, value)


@pytest.mark.benchmark()
def test_bench_get_multi(request, client, pairs, count):
    client.set_multi(pairs)
    benchmark(count, client.get_multi, list(pairs))


@pytest.mark.benchmark()
def test_bench_set_multi(request, client, pairs, count):
    benchmark(count, client.set_multi, pairs)


@pytest.mark.benchmark()
def test_bench_delete(request, client, pairs, count):
    benchmark(count, client.delete, next(pairs))


@pytest.mark.benchmark()
def test_bench_delete_multi(request, client, pairs, count):
    # deleting missing key takes the same work client-side as real keys
    benchmark(count, client.delete_multi, list(pairs.keys()))



import os.path
import socket
import ssl

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--server", action="store", default="localhost", help="memcached server"
    )

    parser.addoption(
        "--port", action="store", default="11211", help="memcached server port"
    )

    parser.addoption(
        "--tls-server", action="store", default="localhost", help="TLS memcached server"
    )

    parser.addoption(
        "--tls-port", action="store", default="11212", help="TLS memcached server port"
    )

    parser.addoption(
        "--size", action="store", default=1024, help="size of data in benchmarks"
    )

    parser.addoption(
        "--count",
        action="store",
        default=10000,
        help="number of iterations to run each benchmark",
    )

    parser.addoption(
        "--keys",
        action="store",
        default=20,
        help="number of keys to use for multi benchmarks",
    )


@pytest.fixture(scope="session")
def host(request):
    return request.config.option.server


@pytest.fixture(scope="session")
def port(request):
    return int(request.config.option.port)


@pytest.fixture(scope="session")
def tls_host(request):
    return request.config.option.tls_server


@pytest.fixture(scope="session")
def tls_port(request):
    return int(request.config.option.tls_port)


@pytest.fixture(scope="session")
def size(request):
    return int(request.config.option.size)


@pytest.fixture(scope="session")
def count(request):
    return int(request.config.option.count)


@pytest.fixture(scope="session")
def keys(request):
    return int(request.config.option.keys)


@pytest.fixture(scope="session")
def pairs(size, keys):
    return {"pymemcache_test:%d" % i: "X" * size for i in range(keys)}


@pytest.fixture(scope="session")
def tls_context():
    return ssl.create_default_context(
        cafile=os.path.join(os.path.dirname(__file__), "certs/ca-root.crt")
    )


def pytest_generate_tests(metafunc):
    if "socket_module" in metafunc.fixturenames:
        socket_modules = [socket]
        try:
            from gevent import socket as gevent_socket  # type: ignore
        except ImportError:
            print("Skipping gevent (not installed)")
        else:
            socket_modules.append(gevent_socket)

        metafunc.parametrize("socket_module", socket_modules)

    if "client_class" in metafunc.fixturenames:
        from pymemcache.client.base import Client, PooledClient
        from pymemcache.client.hash import HashClient

        class HashClientSingle(HashClient):
            def __init__(self, server, *args, **kwargs):
                super().__init__([server], *args, **kwargs)

        metafunc.parametrize("client_class", [Client, PooledClient, HashClientSingle])

    if "key_prefix" in metafunc.fixturenames:
        mark = metafunc.definition.get_closest_marker("parametrize")
        if not mark or "key_prefix" not in mark.args[0]:
            metafunc.parametrize("key_prefix", [b"", b"prefix"])



from pymemcache.client.hash import HashClient
from pymemcache.client.base import Client, PooledClient
from pymemcache.exceptions import MemcacheError, MemcacheUnknownError
from pymemcache import pool

from .test_client import ClientTestMixin, MockSocket
import unittest
import os
import pytest
from unittest import mock
import socket


class TestHashClient(ClientTestMixin, unittest.TestCase):
    def make_client_pool(self, hostname, mock_socket_values, serializer=None, **kwargs):
        mock_client = Client(hostname, serializer=serializer, **kwargs)
        mock_client.sock = MockSocket(mock_socket_values)
        client = PooledClient(hostname, serializer=serializer)
        client.client_pool = pool.ObjectPool(lambda: mock_client)
        return mock_client

    def make_client(self, *mock_socket_values, **kwargs):
        current_port = 11012
        client = HashClient([], **kwargs)
        ip = "127.0.0.1"

        for vals in mock_socket_values:
            s = f"{ip}:{current_port}"
            c = self.make_client_pool((ip, current_port), vals, **kwargs)
            client.clients[s] = c
            client.hasher.add_node(s)
            current_port += 1

        return client

    def make_unix_client(self, sockets, *mock_socket_values, **kwargs):
        client = HashClient([], **kwargs)

        for socket_, vals in zip(sockets, mock_socket_values):
            c = self.make_client_pool(socket_, vals, **kwargs)
            client.clients[socket_] = c
            client.hasher.add_node(socket_)

        return client

    def test_setup_client_without_pooling(self):
        client_class = "pymemcache.client.hash.HashClient.client_class"
        with mock.patch(client_class) as internal_client:
            client = HashClient([], timeout=999, key_prefix="foo_bar_baz")
            client.add_server(("127.0.0.1", "11211"))

        assert internal_client.call_args[0][0] == ("127.0.0.1", "11211")
        kwargs = internal_client.call_args[1]
        assert kwargs["timeout"] == 999
        assert kwargs["key_prefix"] == "foo_bar_baz"

    def test_get_many_unix(self):
        pid = os.getpid()
        sockets = [
            "/tmp/pymemcache.1.%d" % pid,
            "/tmp/pymemcache.2.%d" % pid,
        ]
        client = self.make_unix_client(
            sockets,
            *[
                [
                    b"STORED\r\n",
                    b"VALUE key3 0 6\r\nvalue2\r\nEND\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VALUE key1 0 6\r\nvalue1\r\nEND\r\n",
                ],
            ],
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["/tmp/pymemcache.1.%d" % pid]
            else:
                return client.clients["/tmp/pymemcache.2.%d" % pid]

        client._get_client = get_clients

        result = client.set(b"key1", b"value1", noreply=False)
        result = client.set(b"key3", b"value2", noreply=False)
        result = client.get_many([b"key1", b"key3"])
        assert result == {b"key1": b"value1", b"key3": b"value2"}

    def test_get_many_all_found(self):
        client = self.make_client(
            *[
                [
                    b"STORED\r\n",
                    b"VALUE key3 0 6\r\nvalue2\r\nEND\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VALUE key1 0 6\r\nvalue1\r\nEND\r\n",
                ],
            ]
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["127.0.0.1:11012"]
            else:
                return client.clients["127.0.0.1:11013"]

        client._get_client = get_clients

        result = client.set(b"key1", b"value1", noreply=False)
        result = client.set(b"key3", b"value2", noreply=False)
        result = client.get_many([b"key1", b"key3"])
        assert result == {b"key1": b"value1", b"key3": b"value2"}

    def test_get_many_some_found(self):
        client = self.make_client(
            *[
                [
                    b"END\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VALUE key1 0 6\r\nvalue1\r\nEND\r\n",
                ],
            ]
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["127.0.0.1:11012"]
            else:
                return client.clients["127.0.0.1:11013"]

        client._get_client = get_clients
        result = client.set(b"key1", b"value1", noreply=False)
        result = client.get_many([b"key1", b"key3"])

        assert result == {b"key1": b"value1"}

    def test_get_many_bad_server_data(self):
        client = self.make_client(
            *[
                [
                    b"STORED\r\n",
                    b"VAXLUE key3 0 6\r\nvalue2\r\nEND\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VAXLUE key1 0 6\r\nvalue1\r\nEND\r\n",
                ],
            ]
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["127.0.0.1:11012"]
            else:
                return client.clients["127.0.0.1:11013"]

        client._get_client = get_clients

        with pytest.raises(MemcacheUnknownError):
            client.set(b"key1", b"value1", noreply=False)
            client.set(b"key3", b"value2", noreply=False)
            client.get_many([b"key1", b"key3"])

    def test_get_many_bad_server_data_ignore(self):
        client = self.make_client(
            *[
                [
                    b"STORED\r\n",
                    b"VAXLUE key3 0 6\r\nvalue2\r\nEND\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VAXLUE key1 0 6\r\nvalue1\r\nEND\r\n",
                ],
            ],
            ignore_exc=True,
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["127.0.0.1:11012"]
            else:
                return client.clients["127.0.0.1:11013"]

        client._get_client = get_clients

        client.set(b"key1", b"value1", noreply=False)
        client.set(b"key3", b"value2", noreply=False)
        result = client.get_many([b"key1", b"key3"])
        assert result == {}

    def test_gets_many(self):
        client = self.make_client(
            *[
                [
                    b"STORED\r\n",
                    b"VALUE key3 0 6 1\r\nvalue2\r\nEND\r\n",
                ],
                [
                    b"STORED\r\n",
                    b"VALUE key1 0 6 1\r\nvalue1\r\nEND\r\n",
                ],
            ]
        )

        def get_clients(key):
            if key == b"key3":
                return client.clients["127.0.0.1:11012"]
            else:
                return client.clients["127.0.0.1:11013"]

        client._get_client = get_clients

        assert client.set(b"key1", b"value1", noreply=False) is True
        assert client.set(b"key3", b"value2", noreply=False) is True
        result = client.gets_many([b"key1", b"key3"])
        assert result == {b"key1": (b"value1", b"1"), b"key3": (b"value2", b"1")}

    def test_touch_not_found(self):
        client = self.make_client([b"NOT_FOUND\r\n"])
        result = client.touch(b"key", noreply=False)
        assert result is False

    def test_touch_no_expiry_found(self):
        client = self.make_client([b"TOUCHED\r\n"])
        result = client.touch(b"key", noreply=False)
        assert result is True

    def test_touch_with_expiry_found(self):
        client = self.make_client([b"TOUCHED\r\n"])
        result = client.touch(b"key", 1, noreply=False)
        assert result is True

    def test_close(self):
        client = self.make_client([])
        assert all(c.sock is not None for c in client.clients.values())
        result = client.close()
        assert result is None
        assert all(c.sock is None for c in client.clients.values())

    def test_quit(self):
        client = self.make_client([])
        assert all(c.sock is not None for c in client.clients.values())
        result = client.quit()
        assert result is None
        assert all(c.sock is None for c in client.clients.values())

    def test_no_servers_left(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=True, timeout=1, connect_timeout=1
        )

        hashed_client = client._get_client("foo")
        assert hashed_client is None

    def test_no_servers_left_raise_exception(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=False, timeout=1, connect_timeout=1
        )

        with pytest.raises(MemcacheError) as e:
            client._get_client("foo")

        assert str(e.value) == "All servers seem to be down right now"

    def test_unavailable_servers_zero_retry_raise_exception(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [("example.com", 11211)],
            use_pooling=True,
            ignore_exc=False,
            retry_attempts=0,
            timeout=1,
            connect_timeout=1,
        )

        with pytest.raises(socket.error):
            client.get("foo")

    def test_no_servers_left_with_commands_return_default_value(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=True, timeout=1, connect_timeout=1
        )

        result = client.get("foo")
        assert result is None
        result = client.get("foo", default="default")
        assert result == "default"
        result = client.set("foo", "bar")
        assert result is False

    def test_no_servers_left_return_positional_default(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=True, timeout=1, connect_timeout=1
        )

        # Ensure compatibility with clients that pass the default as a
        # positional argument
        result = client.get("foo", "default")
        assert result == "default"

    def test_no_servers_left_with_set_many(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=True, timeout=1, connect_timeout=1
        )

        result = client.set_many({"foo": "bar"})
        assert result == ["foo"]

    def test_no_servers_left_with_get_many(self):
        from pymemcache.client.hash import HashClient

        client = HashClient(
            [], use_pooling=True, ignore_exc=True, timeout=1, connect_timeout=1
        )

        result = client.get_many(["foo", "bar"])
        assert result == {}

    def test_ignore_exec_set_many(self):
        values = {"key1": "value1", "key2": "value2", "key3": "value3"}

        with pytest.raises(MemcacheUnknownError):
            client = self.make_client(
                *[
                    [b"STORED\r\n", b"UNKNOWN\r\n", b"STORED\r\n"],
                    [b"STORED\r\n", b"UNKNOWN\r\n", b"STORED\r\n"],
                ]
            )
            client.set_many(values, noreply=False)

        client = self.make_client(
            *[
                [b"STORED\r\n", b"UNKNOWN\r\n", b"STORED\r\n"],
            ],
            ignore_exc=True,
        )
        result = client.set_many(values, noreply=False)

        assert len(result) == 0

    def test_noreply_set_many(self):
        values = {"key1": "value1", "key2": "value2", "key3": "value3"}

        client = self.make_client(
            *[
                [b"STORED\r\n", b"NOT_STORED\r\n", b"STORED\r\n"],
            ]
        )
        result = client.set_many(values, noreply=False)
        assert len(result) == 1

        client = self.make_client(
            *[
                [b"STORED\r\n", b"NOT_STORED\r\n", b"STORED\r\n"],
            ]
        )
        result = client.set_many(values, noreply=True)
        assert result == []

    def test_noreply_flush(self):
        client = self.make_client()
        client.flush_all(noreply=True)

    def test_set_many_unix(self):
        values = {"key1": "value1", "key2": "value2", "key3": "value3"}

        pid = os.getpid()
        sockets = ["/tmp/pymemcache.%d" % pid]
        client = self.make_unix_client(
            sockets,
            *[
                [b"STORED\r\n", b"NOT_STORED\r\n", b"STORED\r\n"],
            ],
        )

        result = client.set_many(values, noreply=False)
        assert len(result) == 1

    def test_server_encoding_pooled(self):
        """
        test passed encoding from hash client to pooled clients
        """
        encoding = "utf8"
        from pymemcache.client.hash import HashClient

        hash_client = HashClient(
            [("example.com", 11211)], use_pooling=True, encoding=encoding
        )

        for client in hash_client.clients.values():
            assert client.encoding == encoding

    def test_server_encoding_client(self):
        """
        test passed encoding from hash client to clients
        """
        encoding = "utf8"
        from pymemcache.client.hash import HashClient

        hash_client = HashClient([("example.com", 11211)], encoding=encoding)

        for client in hash_client.clients.values():
            assert client.encoding == encoding

    @mock.patch("pymemcache.client.hash.HashClient.client_class")
    def test_dead_server_comes_back(self, client_patch):
        client = HashClient([], dead_timeout=0, retry_attempts=0)
        client.add_server(("127.0.0.1", 11211))

        test_client = client_patch.return_value
        test_client.server = ("127.0.0.1", 11211)

        test_client.get.side_effect = socket.timeout()
        with pytest.raises(socket.timeout):
            client.get(b"key", noreply=False)
        # Client gets removed because of socket timeout
        assert ("127.0.0.1", 11211) in client._dead_clients

        test_client.get.side_effect = lambda *_, **_kw: "Some value"
        # Client should be retried and brought back
        assert client.get(b"key") == "Some value"
        assert ("127.0.0.1", 11211) not in client._dead_clients

    @mock.patch("pymemcache.client.hash.HashClient.client_class")
    def test_failed_is_retried(self, client_patch):
        client = HashClient([], retry_attempts=1, retry_timeout=0)
        client.add_server(("127.0.0.1", 11211))

        assert client_patch.call_count == 1

        test_client = client_patch.return_value
        test_client.server = ("127.0.0.1", 11211)

        test_client.get.side_effect = socket.timeout()
        with pytest.raises(socket.timeout):
            client.get(b"key", noreply=False)

        test_client.get.side_effect = lambda *_, **_kw: "Some value"
        assert client.get(b"key") == "Some value"

        assert client_patch.call_count == 1

    def test_custom_client(self):
        class MyClient(Client):
            pass

        client = HashClient([])
        client.client_class = MyClient
        client.add_server(("host", 11211))
        assert isinstance(client.clients["host:11211"], MyClient)

    def test_custom_client_with_pooling(self):
        class MyClient(Client):
            pass

        client = HashClient([], use_pooling=True)
        client.client_class = MyClient
        client.add_server(("host", 11211))
        assert isinstance(client.clients["host:11211"], PooledClient)

        pool = client.clients["host:11211"].client_pool
        with pool.get_and_release(destroy_on_fail=True) as c:
            assert isinstance(c, MyClient)

    def test_mixed_inet_and_unix_sockets(self):
        expected = {
            f"/tmp/pymemcache.{os.getpid()}",
            ("127.0.0.1", 11211),
            ("::1", 11211),
        }
        client = HashClient(
            [
                f"/tmp/pymemcache.{os.getpid()}",
                "127.0.0.1",
                "127.0.0.1:11211",
                "[::1]",
                "[::1]:11211",
                ("127.0.0.1", 11211),
                ("::1", 11211),
            ]
        )
        assert expected == {c.server for c in client.clients.values()}

    def test_legacy_add_remove_server_signature(self):
        server = ("127.0.0.1", 11211)
        client = HashClient([])
        assert client.clients == {}
        client.add_server(*server)  # Unpack (host, port) tuple.
        assert ("%s:%s" % server) in client.clients
        client._mark_failed_server(server)
        assert server in client._failed_clients
        client.remove_server(*server)  # Unpack (host, port) tuple.
        assert server in client._dead_clients
        assert server not in client._failed_clients

        # Ensure that server is a string if passing port argument:
        with pytest.raises(TypeError):
            client.add_server(server, server[-1])
        with pytest.raises(TypeError):
            client.remove_server(server, server[-1])

    # TODO: Test failover logic



def murmur3_32(data, seed=0):
    """MurmurHash3 was written by Austin Appleby, and is placed in the
    public domain. The author hereby disclaims copyright to this source
    code."""

    c1 = 0xCC9E2D51
    c2 = 0x1B873593

    length = len(data)
    h1 = seed
    roundedEnd = length & 0xFFFFFFFC  # round down to 4 byte block
    for i in range(0, roundedEnd, 4):
        # little endian load order
        k1 = (
            (ord(data[i]) & 0xFF)
            | ((ord(data[i + 1]) & 0xFF) << 8)
            | ((ord(data[i + 2]) & 0xFF) << 16)
            | (ord(data[i + 3]) << 24)
        )
        k1 *= c1
        k1 = (k1 << 15) | ((k1 & 0xFFFFFFFF) >> 17)  # ROTL32(k1,15)
        k1 *= c2

        h1 ^= k1
        h1 = (h1 << 13) | ((h1 & 0xFFFFFFFF) >> 19)  # ROTL32(h1,13)
        h1 = h1 * 5 + 0xE6546B64

    # tail
    k1 = 0

    val = length & 0x03
    if val == 3:
        k1 = (ord(data[roundedEnd + 2]) & 0xFF) << 16
    # fallthrough
    if val in [2, 3]:
        k1 |= (ord(data[roundedEnd + 1]) & 0xFF) << 8
    # fallthrough
    if val in [1, 2, 3]:
        k1 |= ord(data[roundedEnd]) & 0xFF
        k1 *= c1
        k1 = (k1 << 15) | ((k1 & 0xFFFFFFFF) >> 17)  # ROTL32(k1,15)
        k1 *= c2
        h1 ^= k1

    # finalization
    h1 ^= length

    # fmix(h1)
    h1 ^= (h1 & 0xFFFFFFFF) >> 16
    h1 *= 0x85EBCA6B
    h1 ^= (h1 & 0xFFFFFFFF) >> 13
    h1 *= 0xC2B2AE35
    h1 ^= (h1 & 0xFFFFFFFF) >> 16

    return h1 & 0xFFFFFFFF



# Copyright 2012 Pinterest.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import platform
import socket
from functools import partial
from ssl import SSLContext
from types import ModuleType
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from pymemcache import pool
from pymemcache.exceptions import (
    MemcacheClientError,
    MemcacheIllegalInputError,
    MemcacheServerError,
    MemcacheUnexpectedCloseError,
    MemcacheUnknownCommandError,
    MemcacheUnknownError,
)
from pymemcache.serde import LegacyWrappingSerde

RECV_SIZE = 4096
VALID_STORE_RESULTS = {
    b"set": (b"STORED", b"NOT_STORED"),
    b"add": (b"STORED", b"NOT_STORED"),
    b"replace": (b"STORED", b"NOT_STORED"),
    b"append": (b"STORED", b"NOT_STORED"),
    b"prepend": (b"STORED", b"NOT_STORED"),
    b"cas": (b"STORED", b"EXISTS", b"NOT_FOUND"),
}

SOCKET_KEEPALIVE_SUPPORTED_SYSTEM = {
    "Linux",
}

STORE_RESULTS_VALUE = {
    b"STORED": True,
    b"NOT_STORED": False,
    b"NOT_FOUND": None,
    b"EXISTS": False,
}

ServerSpec = Union[Tuple[str, int], str]
Key = Union[bytes, str]


# Some of the values returned by the "stats" command
# need mapping into native Python types
def _parse_bool_int(value: bytes) -> bool:
    return int(value) != 0


def _parse_bool_string_is_yes(value: bytes) -> bool:
    return value == b"yes"


def _parse_float(value: bytes) -> float:
    return float(value.replace(b":", b"."))


def _parse_hex(value: bytes) -> int:
    return int(value, 8)


STAT_TYPES: Dict[bytes, Callable[[bytes], Any]] = {
    # General stats
    b"version": bytes,
    b"rusage_user": _parse_float,
    b"rusage_system": _parse_float,
    b"hash_is_expanding": _parse_bool_int,
    b"slab_reassign_running": _parse_bool_int,
    # Settings stats
    b"inter": bytes,
    b"growth_factor": float,
    b"stat_key_prefix": bytes,
    b"umask": _parse_hex,
    b"detail_enabled": _parse_bool_int,
    b"cas_enabled": _parse_bool_int,
    b"auth_enabled_sasl": _parse_bool_string_is_yes,
    b"maxconns_fast": _parse_bool_int,
    b"slab_reassign": _parse_bool_int,
    b"slab_automove": _parse_bool_int,
}

# Common helper functions.


def check_key_helper(
    key: Key, allow_unicode_keys: bool, key_prefix: bytes = b""
) -> bytes:
    """Checks key and add key_prefix."""
    if allow_unicode_keys:
        if isinstance(key, str):
            key = key.encode("utf8")
    elif isinstance(key, str):
        try:
            key = key.encode("ascii")
        except (UnicodeEncodeError, UnicodeDecodeError):
            raise MemcacheIllegalInputError("Non-ASCII key: %r" % key)

    key = key_prefix + key
    parts = key.split()

    if len(key) > 250:
        raise MemcacheIllegalInputError("Key is too long: %r" % key)
    # second statement catches leading or trailing whitespace
    elif len(parts) > 1 or (parts and parts[0] != key):
        raise MemcacheIllegalInputError("Key contains whitespace: %r" % key)
    elif b"\00" in key:
        raise MemcacheIllegalInputError("Key contains null: %r" % key)

    return key


def normalize_server_spec(server: ServerSpec) -> ServerSpec:
    if isinstance(server, tuple):
        return server
    if not isinstance(server, str):
        raise ValueError(f"Unsupported server specification: {server!r}")
    if server.startswith("unix:"):
        return server[5:]
    if server.startswith("/"):
        return server
    if ":" not in server or server.endswith("]"):
        host, port = server, 11211
    else:
        parts = server.rsplit(":", 1)
        host, port = parts[0], int(parts[1])
    if host.startswith("["):
        host = host.strip("[]")
    return (host, port)


class KeepaliveOpts:
    """
    A configuration structure to define the socket keepalive.

    This structure must be passed to a client. The client will configure
    its socket keepalive by using the elements of the structure.

    Args:
      idle: The time (in seconds) the connection needs to remain idle
        before TCP starts sending keepalive probes. Should be a positive
        integer most greater than zero.
      intvl: The time (in seconds) between individual keepalive probes.
        Should be a positive integer most greater than zero.
      cnt: The maximum number of keepalive probes TCP should send before
        dropping the connection. Should be a positive integer most greater
        than zero.
    """

    __slots__ = ("idle", "intvl", "cnt")

    def __init__(self, idle: int = 1, intvl: int = 1, cnt: int = 5) -> None:
        if idle < 1:
            raise ValueError("The idle parameter must be greater or equal to 1.")
        self.idle = idle
        if intvl < 1:
            raise ValueError("The intvl parameter must be greater or equal to 1.")
        self.intvl = intvl
        if cnt < 1:
            raise ValueError("The cnt parameter must be greater or equal to 1.")
        self.cnt = cnt


class Client:
    """
    A client for a single memcached server.

    *Server Connection*

     The ``server`` parameter controls how the client connects to the memcached
     server. You can either use a (host, port) tuple for a TCP connection or a
     string containing the path to a UNIX domain socket.

     The ``connect_timeout`` and ``timeout`` parameters can be used to set
     socket timeout values. By default, timeouts are disabled.

     When the ``no_delay`` flag is set, the ``TCP_NODELAY`` socket option will
     also be set. This only applies to TCP-based connections.

     Lastly, the ``socket_module`` allows you to specify an alternate socket
     implementation (such as `gevent.socket`_).

     .. _gevent.socket: http://www.gevent.org/api/gevent.socket.html

    *Keys and Values*

     Keys must have a __str__() method which should return a str with no more
     than 250 ASCII characters and no whitespace or control characters. Unicode
     strings must be encoded (as UTF-8, for example) unless they consist only
     of ASCII characters that are neither whitespace nor control characters.

     Values must have a __str__() method to convert themselves to a byte
     string. Unicode objects can be a problem since str() on a Unicode object
     will attempt to encode it as ASCII (which will fail if the value contains
     code points larger than U+127). You can fix this with a serializer or by
     just calling encode on the string (using UTF-8, for instance).

     If you intend to use anything but str as a value, it is a good idea to use
     a serializer. The pymemcache.serde library has an already implemented
     serializer which pickles and unpickles data.

    *Serialization and Deserialization*

     The constructor takes an optional object, the "serializer/deserializer"
     ("serde"), which is responsible for both serialization and deserialization
     of objects. That object must satisfy the serializer interface by providing
     two methods: `serialize` and `deserialize`. `serialize` takes two
     arguments, a key and a value, and returns a tuple of two elements, the
     serialized value, and an integer in the range 0-65535 (the "flags").
     `deserialize` takes three parameters, a key, value, and flags, and returns
     the deserialized value.

     Here is an example using JSON for non-str values:

     .. code-block:: python

         class JSONSerde(object):
             def serialize(self, key, value):
                 if isinstance(value, str):
                     return value, 1
                 return json.dumps(value), 2

             def deserialize(self, key, value, flags):
                 if flags == 1:
                     return value

                 if flags == 2:
                     return json.loads(value)

                 raise Exception("Unknown flags for value: {1}".format(flags))

    .. note::

     Most write operations allow the caller to provide a ``flags`` value to
     support advanced interaction with the server. This will **override** the
     "flags" value returned by the serializer and should therefore only be
     used when you have a complete understanding of how the value should be
     serialized, stored, and deserialized.

    *Error Handling*

     All of the methods in this class that talk to memcached can throw one of
     the following exceptions:

      * :class:`pymemcache.exceptions.MemcacheUnknownCommandError`
      * :class:`pymemcache.exceptions.MemcacheClientError`
      * :class:`pymemcache.exceptions.MemcacheServerError`
      * :class:`pymemcache.exceptions.MemcacheUnknownError`
      * :class:`pymemcache.exceptions.MemcacheUnexpectedCloseError`
      * :class:`pymemcache.exceptions.MemcacheIllegalInputError`
      * :class:`socket.timeout`
      * :class:`socket.error`

     Instances of this class maintain a persistent connection to memcached
     which is terminated when any of these exceptions are raised. The next
     call to a method on the object will result in a new connection being made
     to memcached.
    """

    def __init__(
        self,
        server: ServerSpec,
        serde=None,
        serializer=None,
        deserializer=None,
        connect_timeout: Optional[float] = None,
        timeout: Optional[float] = None,
        no_delay: bool = False,
        ignore_exc: bool = False,
        socket_module: ModuleType = socket,
        socket_keepalive: Optional[KeepaliveOpts] = None,
        key_prefix: bytes = b"",
        default_noreply: bool = True,
        allow_unicode_keys: bool = False,
        encoding: str = "ascii",
        tls_context: Optional[SSLContext] = None,
    ):
        """
        Constructor.

        Args:
          server: tuple(hostname, port) or string containing a UNIX socket path.
          serde: optional serializer object, see notes in the class docs.
          serializer: deprecated serialization function
          deserializer: deprecated deserialization function
          connect_timeout: optional float, seconds to wait for a connection to
            the memcached server. Defaults to "forever" (uses the underlying
            default socket timeout, which can be very long).
          timeout: optional float, seconds to wait for send or recv calls on
            the socket connected to memcached. Defaults to "forever" (uses the
            underlying default socket timeout, which can be very long).
          no_delay: optional bool, set the TCP_NODELAY flag, which may help
            with performance in some cases. Defaults to False.
          ignore_exc: optional bool, True to cause the "get", "gets",
            "get_many" and "gets_many" calls to treat any errors as cache
            misses. Defaults to False.
          socket_module: socket module to use, e.g. gevent.socket. Defaults to
            the standard library's socket module.
          socket_keepalive: Activate the socket keepalive feature by passing
            a KeepaliveOpts structure in this parameter. Disabled by default
            (None). This feature is only supported on Linux platforms.
          key_prefix: Prefix of key. You can use this as namespace. Defaults
            to b''.
          default_noreply: bool, the default value for 'noreply' as passed to
            store commands (except from cas, incr, and decr, which default to
            False).
          allow_unicode_keys: bool, support unicode (utf8) keys
          encoding: optional str, controls data encoding (defaults to 'ascii').

        Notes:
          The constructor does not make a connection to memcached. The first
          call to a method on the object will do that.
        """
        self.server = normalize_server_spec(server)
        self.serde = serde or LegacyWrappingSerde(serializer, deserializer)
        self.connect_timeout = connect_timeout
        self.timeout = timeout
        self.no_delay = no_delay
        self.ignore_exc = ignore_exc
        self.socket_module = socket_module
        self.socket_keepalive = socket_keepalive
        user_system = platform.system()
        if self.socket_keepalive is not None:
            if user_system not in SOCKET_KEEPALIVE_SUPPORTED_SYSTEM:
                raise SystemError(
                    "Pymemcache's socket keepalive mechanism doesn't "
                    "support your system ({user_system}). If "
                    "you see this message it mean that you tried to "
                    "configure your socket keepalive on an unsupported "
                    "system. To fix the problem pass `socket_"
                    "keepalive=False` or use a supported system. "
                    "Supported systems are: {systems}".format(
                        user_system=user_system,
                        systems=", ".join(sorted(SOCKET_KEEPALIVE_SUPPORTED_SYSTEM)),
                    )
                )
            if not isinstance(self.socket_keepalive, KeepaliveOpts):
                raise ValueError(
                    "Unsupported keepalive options. If you see this message "
                    "it means that you passed an unsupported object within "
                    "the param `socket_keepalive`. To fix it "
                    "please instantiate and pass to socket_keepalive a "
                    "KeepaliveOpts object. That's the only supported type "
                    "of structure."
                )
        self.sock: Optional[socket.socket] = None
        if isinstance(key_prefix, str):
            key_prefix = key_prefix.encode("ascii")
        if not isinstance(key_prefix, bytes):
            raise TypeError("key_prefix should be bytes.")
        self.key_prefix = key_prefix
        self.default_noreply = default_noreply
        self.allow_unicode_keys = allow_unicode_keys
        self.encoding = encoding
        self.tls_context = tls_context

    def check_key(self, key: Key, key_prefix: bytes) -> bytes:
        """Checks key and add key_prefix."""
        return check_key_helper(
            key, allow_unicode_keys=self.allow_unicode_keys, key_prefix=key_prefix
        )

    def _connect(self) -> None:
        self.close()

        s = self.socket_module

        if not isinstance(self.server, tuple):
            sockaddr = self.server
            sock = s.socket(s.AF_UNIX, s.SOCK_STREAM)
        else:
            sock = None
            error = None
            host, port = self.server
            info = s.getaddrinfo(host, port, s.AF_UNSPEC, s.SOCK_STREAM, s.IPPROTO_TCP)
            for family, socktype, proto, _, sockaddr in info:
                try:
                    sock = s.socket(family, socktype, proto)
                    if self.no_delay:
                        sock.setsockopt(s.IPPROTO_TCP, s.TCP_NODELAY, 1)
                    if self.tls_context:
                        context = self.tls_context
                        sock = context.wrap_socket(sock, server_hostname=host)
                except Exception as e:
                    error = e
                    if sock is not None:
                        sock.close()
                        sock = None
                else:
                    break

            if error is not None:
                raise error

        try:
            sock.settimeout(self.connect_timeout)
            if self.socket_keepalive is not None:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, self.socket_keepalive.idle
                )
                sock.setsockopt(
                    socket.IPPROTO_TCP,
                    socket.TCP_KEEPINTVL,
                    self.socket_keepalive.intvl,
                )
                sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPCNT, self.socket_keepalive.cnt
                )
            sock.connect(sockaddr)
            sock.settimeout(self.timeout)
        except Exception:
            sock.close()
            raise

        self.sock = sock

    def close(self) -> None:
        """Close the connection to memcached, if it is open. The next call to a
        method that requires a connection will re-open it."""
        if self.sock is not None:
            try:
                self.sock.close()
            except Exception:
                pass
            finally:
                self.sock = None

    disconnect_all = close

    def set(
        self,
        key: Key,
        value: Any,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ) -> Optional[bool]:
        """
        The memcached "set" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                flags

        Returns:
          If no exception is raised, always returns True. If an exception is
          raised, the set may or may not have occurred. If noreply is True,
          then a successful return does not guarantee a successful set.
        """
        if noreply is None:
            noreply = self.default_noreply
        # Optional because _store_cmd lookup in STORE_RESULTS_VALUE can return None in some cases.
        # TODO: refactor to fix
        return self._store_cmd(b"set", {key: value}, expire, noreply, flags=flags)[key]

    def set_many(
        self,
        values: Dict[Key, Any],
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ) -> List[Key]:
        """
        A convenience function for setting multiple values.

        Args:
          values: dict(str, str), a dict of keys and values, see class docs
                  for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                 flags

        Returns:
          Returns a list of keys that failed to be inserted.
          If noreply is True, always returns empty list.
        """
        if noreply is None:
            noreply = self.default_noreply
        result = self._store_cmd(b"set", values, expire, noreply, flags=flags)
        return [k for k, v in result.items() if not v]

    set_multi = set_many

    def add(
        self,
        key: Key,
        value: Any,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ) -> bool:
        """
        The memcached "add" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                  flags

        Returns:
          If ``noreply`` is True (or if it is unset and ``self.default_noreply``
          is True), the return value is always True. Otherwise the return value
          is True if the value was stored, and False if it was not (because
          the key already existed).
        """
        if noreply is None:
            noreply = self.default_noreply
        response = self._store_cmd(b"add", {key: value}, expire, noreply, flags=flags)[
            key
        ]
        # For typing, can only be None, if cas command
        assert response is not None
        return response

    def replace(
        self,
        key: Key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ) -> bool:
        """
        The memcached "replace" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                flags

        Returns:
          If ``noreply`` is True (or if it is unset and ``self.default_noreply``
          is True), the return value is always True. Otherwise returns True if
          the value was stored and False if it wasn't (because the key didn't
          already exist).
        """
        if noreply is None:
            noreply = self.default_noreply
        response = self._store_cmd(
            b"replace", {key: value}, expire, noreply, flags=flags
        )[key]
        # for typing
        assert response is not None
        return response

    def append(
        self,
        key: Key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ) -> bool:
        """
        The memcached "append" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                flags

        Returns:
          True.
        """
        if noreply is None:
            noreply = self.default_noreply
        response = self._store_cmd(
            b"append", {key: value}, expire, noreply, flags=flags
        )[key]
        # For typing
        assert response is not None
        return response

    def prepend(
        self,
        key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        """
        The memcached "prepend" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).
          flags: optional int, arbitrary bit field used for server-specific
                flags

        Returns:
          True.
        """
        if noreply is None:
            noreply = self.default_noreply
        return self._store_cmd(b"prepend", {key: value}, expire, noreply, flags=flags)[
            key
        ]

    def cas(
        self,
        key,
        value,
        cas,
        expire: int = 0,
        noreply=False,
        flags: Optional[int] = None,
    ) -> Optional[bool]:
        """
        The memcached "cas" command.

        Args:
          key: str, see class docs for details.
          value: str, see class docs for details.
          cas: int or str that only contains the characters '0'-'9'.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, False to wait for the reply (the default).
          flags: optional int, arbitrary bit field used for server-specific
                flags

        Returns:
          If ``noreply`` is True (or if it is unset and ``self.default_noreply``
          is True), the return value is always True. Otherwise returns None if
          the key didn't exist, False if it existed but had a different cas
          value and True if it existed and was changed.
        """
        cas = self._check_cas(cas)
        return self._store_cmd(
            b"cas", {key: value}, expire, noreply, flags=flags, cas=cas
        )[key]

    def get(self, key: Key, default: Optional[Any] = None) -> Any:
        """
        The memcached "get" command, but only for one key, as a convenience.

        Args:
          key: str, see class docs for details.
          default: value that will be returned if the key was not found.

        Returns:
          The value for the key, or default if the key wasn't found.
        """
        return self._fetch_cmd(b"get", [key], False, key_prefix=self.key_prefix).get(
            key, default
        )

    def get_many(self, keys: Iterable[Key]) -> Dict[Key, Any]:
        """
        The memcached "get" command.

        Args:
          keys: list(str), see class docs for details.

        Returns:
          A dict in which the keys are elements of the "keys" argument list
          and the values are values from the cache. The dict may contain all,
          some or none of the given keys.
        """
        if not keys:
            return {}

        return self._fetch_cmd(b"get", keys, False, key_prefix=self.key_prefix)

    get_multi = get_many

    def gets(
        self, key: Key, default: Any = None, cas_default: Any = None
    ) -> Tuple[Any, Any]:
        """
        The memcached "gets" command for one key, as a convenience.

        Args:
          key: str, see class docs for details.
          default: value that will be returned if the key was not found.
          cas_default: same behaviour as default argument.

        Returns:
          A tuple of (value, cas)
          or (default, cas_defaults) if the key was not found.
        """
        defaults = (default, cas_default)
        return self._fetch_cmd(b"gets", [key], True, key_prefix=self.key_prefix).get(
            key, defaults
        )

    def gets_many(self, keys: Iterable[Key]) -> Dict[Key, Tuple[Any, Any]]:
        """
        The memcached "gets" command.

        Args:
          keys: list(str), see class docs for details.

        Returns:
          A dict in which the keys are elements of the "keys" argument list and
          the values are tuples of (value, cas) from the cache. The dict may
          contain all, some or none of the given keys.
        """
        if not keys:
            return {}

        return self._fetch_cmd(b"gets", keys, True, key_prefix=self.key_prefix)

    def delete(self, key: Key, noreply: Optional[bool] = None) -> bool:
        """
        The memcached "delete" command.

        Args:
          key: str, see class docs for details.
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).

        Returns:
          If ``noreply`` is True (or if it is unset and ``self.default_noreply``
          is True), the return value is always True. Otherwise returns True if
          the key was deleted, and False if it wasn't found.
        """
        if noreply is None:
            noreply = self.default_noreply
        cmd = b"delete " + self.check_key(key, self.key_prefix)
        if noreply:
            cmd += b" noreply"
        cmd += b"\r\n"
        results = self._misc_cmd([cmd], b"delete", noreply)
        if noreply:
            return True
        return results[0] == b"DELETED"

    def delete_many(self, keys: Iterable[Key], noreply: Optional[bool] = None) -> bool:
        """
        A convenience function to delete multiple keys.

        Args:
          keys: list(str), the list of keys to delete.
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).

        Returns:
          True. If an exception is raised then all, some or none of the keys
          may have been deleted. Otherwise all the keys have been sent to
          memcache for deletion and if noreply is False, they have been
          acknowledged by memcache.
        """
        if not keys:
            return True

        if noreply is None:
            noreply = self.default_noreply

        cmds = []
        for key in keys:
            cmds.append(
                b"delete "
                + self.check_key(key, self.key_prefix)
                + (b" noreply" if noreply else b"")
                + b"\r\n"
            )
        self._misc_cmd(cmds, b"delete", noreply)
        return True

    delete_multi = delete_many

    def incr(
        self, key: Key, value: int, noreply: Optional[bool] = False
    ) -> Optional[int]:
        """
        The memcached "incr" command.

        Args:
          key: str, see class docs for details.
          value: int, the amount by which to increment the value.
          noreply: optional bool, False to wait for the reply (the default).

        Returns:
          If noreply is True, always returns None. Otherwise returns the new
          value of the key, or None if the key wasn't found.
        """
        key = self.check_key(key, self.key_prefix)
        val = self._check_integer(value, "value")
        cmd = b"incr " + key + b" " + val
        if noreply:
            cmd += b" noreply"
        cmd += b"\r\n"
        results = self._misc_cmd([cmd], b"incr", noreply)
        if noreply:
            return None
        if results[0] == b"NOT_FOUND":
            return None
        return int(results[0])

    def decr(
        self, key: Key, value: int, noreply: Optional[bool] = False
    ) -> Optional[int]:
        """
        The memcached "decr" command.

        Args:
          key: str, see class docs for details.
          value: int, the amount by which to decrement the value.
          noreply: optional bool, False to wait for the reply (the default).

        Returns:
          If noreply is True, always returns None. Otherwise returns the new
          value of the key, or None if the key wasn't found.
        """
        key = self.check_key(key, self.key_prefix)
        val = self._check_integer(value, "value")
        cmd = b"decr " + key + b" " + val
        if noreply:
            cmd += b" noreply"
        cmd += b"\r\n"
        results = self._misc_cmd([cmd], b"decr", noreply)
        if noreply:
            return None
        if results[0] == b"NOT_FOUND":
            return None
        return int(results[0])

    def touch(self, key: Key, expire: int = 0, noreply: Optional[bool] = None) -> bool:
        """
        The memcached "touch" command.

        Args:
          key: str, see class docs for details.
          expire: optional int, number of seconds until the item is expired
                  from the cache, or zero for no expiry (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).

        Returns:
          True if the expiration time was updated, False if the key wasn't
          found.
        """
        if noreply is None:
            noreply = self.default_noreply
        key = self.check_key(key, self.key_prefix)
        expire_bytes = self._check_integer(expire, "expire")
        cmd = b"touch " + key + b" " + expire_bytes
        if noreply:
            cmd += b" noreply"
        cmd += b"\r\n"
        results = self._misc_cmd([cmd], b"touch", noreply)
        if noreply:
            return True
        return results[0] == b"TOUCHED"

    def stats(self, *args):
        """
        The memcached "stats" command.

        The returned keys depend on what the "stats" command returns.
        A best effort is made to convert values to appropriate Python
        types, defaulting to strings when a conversion cannot be made.

        Args:
          *arg: extra string arguments to the "stats" command. See the
                memcached protocol documentation for more information.

        Returns:
          A dict of the returned stats.
        """
        result = self._fetch_cmd(b"stats", args, False)

        for key, value in result.items():
            converter = STAT_TYPES.get(key, int)
            try:
                result[key] = converter(value)
            except Exception:
                pass

        return result

    def cache_memlimit(self, memlimit) -> bool:
        """
        The memcached "cache_memlimit" command.

        Args:
          memlimit: int, the number of megabytes to set as the new cache memory
                    limit.

        Returns:
          If no exception is raised, always returns True.
        """
        memlimit = self._check_integer(memlimit, "memlimit")
        self._fetch_cmd(b"cache_memlimit", [memlimit], False)
        return True

    def version(self) -> bytes:
        """
        The memcached "version" command.

        Returns:
            A string of the memcached version.
        """
        cmd = b"version\r\n"
        results = self._misc_cmd([cmd], b"version", False)
        before, _, after = results[0].partition(b" ")

        if before != b"VERSION":
            raise MemcacheUnknownError(f"Received unexpected response: {results[0]!r}")
        return after

    def raw_command(
        self, command: Union[str, bytes], end_tokens: Union[str, bytes] = "\r\n"
    ) -> bytes:
        """
        Sends an arbitrary command to the server and parses the response until a
        specified token is encountered.

        Args:
            command: str|bytes: The command to send.
            end_tokens: str|bytes: The token expected at the end of the
                response. If the `end_token` is not found, the client will wait
                until the timeout specified in the constructor.

        Returns:
            The response from the server, with the `end_token` removed.
        """
        encoding = "utf8" if self.allow_unicode_keys else "ascii"
        command = command.encode(encoding) if isinstance(command, str) else command
        end_tokens = (
            end_tokens.encode(encoding) if isinstance(end_tokens, str) else end_tokens
        )
        return self._misc_cmd([b"" + command + b"\r\n"], command, False, end_tokens)[0]

    def flush_all(self, delay: int = 0, noreply: Optional[bool] = None) -> bool:
        """
        The memcached "flush_all" command.

        Args:
          delay: optional int, the number of seconds to wait before flushing,
                 or zero to flush immediately (the default).
          noreply: optional bool, True to not wait for the reply (defaults to
                   self.default_noreply).

        Returns:
          True.
        """
        if noreply is None:
            noreply = self.default_noreply
        delay_bytes = self._check_integer(delay, "delay")
        cmd = b"flush_all " + delay_bytes
        if noreply:
            cmd += b" noreply"
        cmd += b"\r\n"
        results = self._misc_cmd([cmd], b"flush_all", noreply)
        if noreply:
            return True
        return results[0] == b"OK"

    def quit(self) -> None:
        """
        The memcached "quit" command.

        This will close the connection with memcached. Calling any other
        method on this object will re-open the connection, so this object can
        be re-used after quit.
        """
        cmd = b"quit\r\n"
        self._misc_cmd([cmd], b"quit", True)
        self.close()

    def shutdown(self, graceful: bool = False) -> None:
        """
        The memcached "shutdown" command.

        This will request shutdown and eventual termination of the server,
        optionally preceded by a graceful stop of memcached's internal state
        machine. Note that the server needs to have been started with the
        shutdown protocol command enabled with the --enable-shutdown flag.

        Args:
          graceful: optional bool, True to request a graceful shutdown with
                    SIGUSR1 (defaults to False, i.e. SIGINT shutdown).
        """
        cmd = b"shutdown"
        if graceful:
            cmd += b" graceful"
        cmd += b"\r\n"

        # The shutdown command raises a server-side error if the shutdown
        # protocol command is not enabled. Otherwise, a successful shutdown
        # is expected to close the remote end of the transport.
        try:
            self._misc_cmd([cmd], b"shutdown", False)
        except MemcacheUnexpectedCloseError:
            pass

    def _raise_errors(self, line: bytes, name: bytes) -> None:
        if line.startswith(b"ERROR"):
            raise MemcacheUnknownCommandError(name)

        if line.startswith(b"CLIENT_ERROR"):
            error = line[line.find(b" ") + 1 :]
            raise MemcacheClientError(error)

        if line.startswith(b"SERVER_ERROR"):
            error = line[line.find(b" ") + 1 :]
            raise MemcacheServerError(error)

    def _check_integer(self, value: int, name: str) -> bytes:
        """Check that a value is an integer and encode it as a binary string"""
        if not isinstance(value, int):
            raise MemcacheIllegalInputError(
                f"{name} must be integer, got bad value: {value!r}"
            )

        return str(value).encode(self.encoding)

    def _check_cas(self, cas: Union[int, str, bytes]) -> bytes:
        """Check that a value is a valid input for 'cas' -- either an int or a
        string containing only 0-9

        The value will be (re)encoded so that we can accept strings or bytes.
        """
        # convert non-binary values to binary
        if isinstance(cas, (int, str)):
            try:
                cas = str(cas).encode(self.encoding)
            except UnicodeEncodeError:
                raise MemcacheIllegalInputError("non-ASCII cas value: %r" % cas)
        elif not isinstance(cas, bytes):
            raise MemcacheIllegalInputError(
                "cas must be integer, string, or bytes, got bad value: %r" % cas
            )

        if not cas.isdigit():
            raise MemcacheIllegalInputError(
                "cas must only contain values in 0-9, got bad value: %r" % cas
            )

        return cas

    def _extract_value(
        self,
        expect_cas: bool,
        line: bytes,
        buf: bytes,
        remapped_keys: Dict[bytes, Key],
        prefixed_keys: List[bytes],
    ) -> Tuple[Key, Union[Any, Tuple[Any, bytes]], bytes]:
        """
        This function is abstracted from _fetch_cmd to support different ways
        of value extraction. In order to use this feature, _extract_value needs
        to be overridden in the subclass.
        """
        if expect_cas:
            _, key, flags, size, cas = line.split()
        else:
            try:
                _, key, flags, size = line.split()
            except Exception as e:
                raise ValueError(f"Unable to parse line {line!r}: {e}")

        value = None
        try:
            # For typing
            assert self.sock is not None

            buf, value = _readvalue(self.sock, buf, int(size))
        except MemcacheUnexpectedCloseError:
            self.close()
            raise
        original_key = remapped_keys[key]
        value = self.serde.deserialize(original_key, value, int(flags))

        if expect_cas:
            return original_key, (value, cas), buf
        else:
            return original_key, value, buf

    def _fetch_cmd(
        self,
        name: bytes,
        keys: Iterable[Key],
        expect_cas: bool,
        key_prefix: bytes = b"",
    ) -> Dict[Key, Any]:
        prefixed_keys = [self.check_key(k, key_prefix=key_prefix) for k in keys]
        remapped_keys = dict(zip(prefixed_keys, keys))

        # It is important for all keys to be listed in their original order.
        cmd = name
        if prefixed_keys:
            cmd += b" " + b" ".join(prefixed_keys)
        cmd += b"\r\n"

        try:
            if self.sock is None:
                self._connect()

                # For typing
                assert self.sock is not None

            self.sock.sendall(cmd)

            buf = b""
            line = None
            result: Dict[Key, Any] = {}
            while True:
                try:
                    buf, line = _readline(self.sock, buf)
                except MemcacheUnexpectedCloseError:
                    self.close()
                    raise
                self._raise_errors(line, name)
                if line == b"END" or line == b"OK":
                    return result
                elif line.startswith(b"VALUE"):
                    key, value, buf = self._extract_value(
                        expect_cas, line, buf, remapped_keys, prefixed_keys
                    )
                    result[key] = value
                elif name == b"stats" and line.startswith(b"STAT"):
                    key_value = line.split()
                    result[key_value[1]] = key_value[2] if len(key_value) > 2 else b""
                elif name == b"stats" and line.startswith(b"ITEM"):
                    # For 'stats cachedump' commands
                    key_value = line.split()
                    result[key_value[1]] = b" ".join(key_value[2:])
                else:
                    raise MemcacheUnknownError(line[:32])
        except Exception:
            self.close()
            if self.ignore_exc:
                return {}
            raise

    def _store_cmd(
        self,
        name: bytes,
        values: Dict[Key, Any],
        expire: int,
        noreply: bool,
        flags: Optional[int] = None,
        cas: Optional[bytes] = None,
    ) -> Dict[Key, Optional[bool]]:
        cmds = []
        keys = []

        extra = b""
        if cas is not None:
            extra += b" " + cas
        if noreply:
            extra += b" noreply"
        expire_bytes = self._check_integer(expire, "expire")

        for key, data in values.items():
            # must be able to reliably map responses back to the original order
            keys.append(key)

            key = self.check_key(key, self.key_prefix)
            data, data_flags = self.serde.serialize(key, data)

            # If 'flags' was explicitly provided, it overrides the value
            # returned by the serializer.
            if flags is not None:
                data_flags = flags

            if not isinstance(data, bytes):
                try:
                    data = str(data).encode(self.encoding)
                except UnicodeEncodeError as e:
                    raise MemcacheIllegalInputError(
                        "Data values must be binary-safe: %s" % e
                    )

            cmds.append(
                name
                + b" "
                + key
                + b" "
                + str(data_flags).encode(self.encoding)
                + b" "
                + expire_bytes
                + b" "
                + str(len(data)).encode(self.encoding)
                + extra
                + b"\r\n"
                + data
                + b"\r\n"
            )

        if self.sock is None:
            self._connect()

            # For typing
            assert self.sock is not None

        try:
            self.sock.sendall(b"".join(cmds))
            if noreply:
                return {k: True for k in keys}

            results = {}
            buf = b""
            line = None
            for key in keys:
                try:
                    buf, line = _readline(self.sock, buf)
                except MemcacheUnexpectedCloseError:
                    self.close()
                    raise
                self._raise_errors(line, name)

                if line in VALID_STORE_RESULTS[name]:
                    results[key] = STORE_RESULTS_VALUE[line]
                else:
                    raise MemcacheUnknownError(line[:32])
            return results
        except Exception:
            self.close()
            raise

    def _misc_cmd(
        self,
        cmds: Iterable[bytes],
        cmd_name: bytes,
        noreply: Optional[bool],
        end_tokens=None,
    ) -> List[bytes]:

        # If no end_tokens have been given, just assume standard memcached
        # operations, which end in "\r\n", use regular code for that.
        _reader: Callable[[socket.socket, bytes], Tuple[bytes, bytes]]
        if end_tokens:
            _reader = partial(_readsegment, end_tokens=end_tokens)
        else:
            _reader = _readline

        if self.sock is None:
            self._connect()

            # For typing
            assert self.sock is not None

        try:
            self.sock.sendall(b"".join(cmds))

            if noreply:
                return []

            results = []
            buf = b""
            line = None
            for cmd in cmds:
                try:
                    buf, line = _reader(self.sock, buf)
                except MemcacheUnexpectedCloseError:
                    self.close()
                    raise
                self._raise_errors(line, cmd_name)
                results.append(line)
            return results

        except Exception:
            self.close()
            raise

    def __setitem__(self, key: Key, value):
        self.set(key, value, noreply=True)

    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError
        return value

    def __delitem__(self, key):
        self.delete(key, noreply=True)


class PooledClient:
    """A thread-safe pool of clients (with the same client api).

    Args:
      max_pool_size: maximum pool size to use (going above this amount
                     triggers a runtime error), by default this is 2147483648L
                     when not provided (or none).
      pool_idle_timeout: pooled connections are discarded if they have been
                         unused for this many seconds. A value of 0 indicates
                         that pooled connections are never discarded.
      lock_generator: a callback/type that takes no arguments that will
                      be called to create a lock or semaphore that can
                      protect the pool from concurrent access (for example a
                      eventlet lock or semaphore could be used instead)

    Further arguments are interpreted as for :py:class:`.Client` constructor.

    Note: if `serde` is given, the same object will be used for *all* clients
    in the pool. Your serde object must therefore be thread-safe.
    """

    #: :class:`Client` class used to create new clients
    client_class = Client

    def __init__(
        self,
        server: ServerSpec,
        serde=None,
        serializer=None,
        deserializer=None,
        connect_timeout=None,
        timeout=None,
        no_delay=False,
        ignore_exc=False,
        socket_module=socket,
        socket_keepalive=None,
        key_prefix=b"",
        max_pool_size=None,
        pool_idle_timeout=0,
        lock_generator=None,
        default_noreply: bool = True,
        allow_unicode_keys=False,
        encoding="ascii",
        tls_context=None,
    ):
        self.server = normalize_server_spec(server)
        self.serde = serde or LegacyWrappingSerde(serializer, deserializer)
        self.connect_timeout = connect_timeout
        self.timeout = timeout
        self.no_delay = no_delay
        self.ignore_exc = ignore_exc
        self.socket_module = socket_module
        self.socket_keepalive = socket_keepalive
        self.default_noreply = default_noreply
        self.allow_unicode_keys = allow_unicode_keys
        if isinstance(key_prefix, str):
            key_prefix = key_prefix.encode("ascii")
        if not isinstance(key_prefix, bytes):
            raise TypeError("key_prefix should be bytes.")
        self.key_prefix = key_prefix
        self.client_pool = pool.ObjectPool(
            self._create_client,
            after_remove=lambda client: client.close(),
            max_size=max_pool_size,
            idle_timeout=pool_idle_timeout,
            lock_generator=lock_generator,
        )
        self.encoding = encoding
        self.tls_context = tls_context

    def check_key(self, key: Key) -> bytes:
        """Checks key and add key_prefix."""
        return check_key_helper(
            key, allow_unicode_keys=self.allow_unicode_keys, key_prefix=self.key_prefix
        )

    def _create_client(self) -> Client:
        return self.client_class(
            self.server,
            serde=self.serde,
            connect_timeout=self.connect_timeout,
            timeout=self.timeout,
            no_delay=self.no_delay,
            # We need to know when it fails *always* so that we
            # can remove/destroy it from the pool...
            ignore_exc=False,
            socket_module=self.socket_module,
            socket_keepalive=self.socket_keepalive,
            key_prefix=self.key_prefix,
            default_noreply=self.default_noreply,
            allow_unicode_keys=self.allow_unicode_keys,
            tls_context=self.tls_context,
        )

    def close(self) -> None:
        self.client_pool.clear()

    disconnect_all = close

    def set(
        self,
        key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.set(key, value, expire=expire, noreply=noreply, flags=flags)

    def set_many(
        self,
        values,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.set_many(values, expire=expire, noreply=noreply, flags=flags)

    set_multi = set_many

    def replace(
        self,
        key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.replace(
                key, value, expire=expire, noreply=noreply, flags=flags
            )

    def append(
        self,
        key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.append(
                key, value, expire=expire, noreply=noreply, flags=flags
            )

    def prepend(
        self,
        key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.prepend(
                key, value, expire=expire, noreply=noreply, flags=flags
            )

    def cas(
        self,
        key,
        value,
        cas,
        expire: int = 0,
        noreply=False,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.cas(
                key, value, cas, expire=expire, noreply=noreply, flags=flags
            )

    def get(self, key: Key, default: Any = None) -> Any:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                return client.get(key, default)
            except Exception:
                if self.ignore_exc:
                    return default
                else:
                    raise

    def get_many(self, keys: Iterable[Key]) -> Dict[Key, Any]:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                return client.get_many(keys)
            except Exception:
                if self.ignore_exc:
                    return {}
                else:
                    raise

    get_multi = get_many

    def gets(self, key: Key) -> Tuple[Any, Any]:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                return client.gets(key)
            except Exception:
                if self.ignore_exc:
                    return (None, None)
                else:
                    raise

    def gets_many(self, keys: Iterable[Key]) -> Dict[Key, Tuple[Any, Any]]:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                return client.gets_many(keys)
            except Exception:
                if self.ignore_exc:
                    return {}
                else:
                    raise

    def delete(self, key: Key, noreply: Optional[bool] = None) -> bool:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.delete(key, noreply=noreply)

    def delete_many(self, keys: Iterable[Key], noreply: Optional[bool] = None) -> bool:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.delete_many(keys, noreply=noreply)

    delete_multi = delete_many

    def add(
        self,
        key: Key,
        value,
        expire: int = 0,
        noreply: Optional[bool] = None,
        flags: Optional[int] = None,
    ):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.add(key, value, expire=expire, noreply=noreply, flags=flags)

    def incr(self, key: Key, value, noreply=False):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.incr(key, value, noreply=noreply)

    def decr(self, key: Key, value, noreply=False):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.decr(key, value, noreply=noreply)

    def touch(self, key: Key, expire: int = 0, noreply=None):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.touch(key, expire=expire, noreply=noreply)

    def stats(self, *args):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                return client.stats(*args)
            except Exception:
                if self.ignore_exc:
                    return {}
                else:
                    raise

    def version(self) -> bytes:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.version()

    def flush_all(self, delay=0, noreply=None) -> bool:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.flush_all(delay=delay, noreply=noreply)

    def quit(self) -> None:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            try:
                client.quit()
            finally:
                self.client_pool.destroy(client)

    def shutdown(self, graceful: bool = False) -> None:
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            client.shutdown(graceful)

    def raw_command(self, command, end_tokens=b"\r\n"):
        with self.client_pool.get_and_release(destroy_on_fail=True) as client:
            return client.raw_command(command, end_tokens)

    def __setitem__(self, key: Key, value):
        self.set(key, value, noreply=True)

    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError
        return value

    def __delitem__(self, key):
        self.delete(key, noreply=True)


def _readline(sock: socket.socket, buf: bytes) -> Tuple[bytes, bytes]:
    """Read line of text from the socket.

    Read a line of text (delimited by "\r\n") from the socket, and
    return that line along with any trailing characters read from the
    socket.

    Args:
        sock: Socket object, should be connected.
        buf: Bytes, zero or more characters, returned from an earlier
            call to _readline or _readvalue (pass an empty byte string on the
            first call).

    Returns:
      A tuple of (buf, line) where line is the full line read from the
      socket (minus the "\r\n" characters) and buf is any trailing
      characters read after the "\r\n" was found (which may be an empty
      byte string).

    """
    chunks: List[bytes] = []
    last_char = b""

    while True:
        # We're reading in chunks, so "\r\n" could appear in one chunk,
        # or across the boundary of two chunks, so we check for both
        # cases.

        # This case must appear first, since the buffer could have
        # later \r\n characters in it and we want to get the first \r\n.
        if last_char == b"\r" and buf[0:1] == b"\n":
            # Strip the last character from the last chunk.
            chunks[-1] = chunks[-1][:-1]
            return buf[1:], b"".join(chunks)
        else:
            token_pos = buf.find(b"\r\n")
            if token_pos != -1:
                # Note: 2 == len(b"\r\n")
                before, after = buf[:token_pos], buf[token_pos + 2 :]
                chunks.append(before)
                return after, b"".join(chunks)

        if buf:
            chunks.append(buf)
            last_char = buf[-1:]

        buf = _recv(sock, RECV_SIZE)
        if not buf:
            raise MemcacheUnexpectedCloseError()


def _readvalue(sock: socket.socket, buf: bytes, size: int):
    """Read specified amount of bytes from the socket.

    Read size bytes, followed by the "\r\n" characters, from the socket,
    and return those bytes and any trailing bytes read after the "\r\n".

    Args:
        sock: Socket object, should be connected.
        buf: String, zero or more characters, returned from an earlier
            call to _readline or _readvalue (pass an empty string on the
            first call).
        size: Integer, number of bytes to read from the socket.

    Returns:
      A tuple of (buf, value) where value is the bytes read from the
      socket (there will be exactly size bytes) and buf is trailing
      characters read after the "\r\n" following the bytes (but not
      including the \r\n).

    """
    chunks = []
    rlen = size + 2
    while rlen - len(buf) > 0:
        if buf:
            rlen -= len(buf)
            chunks.append(buf)
        buf = _recv(sock, RECV_SIZE)
        if not buf:
            raise MemcacheUnexpectedCloseError()

    # Now we need to remove the \r\n from the end. There are two cases we care
    # about: the \r\n is all in the last buffer, or only the \n is in the last
    # buffer, and we need to remove the \r from the penultimate buffer.

    if rlen == 1:
        # replace the last chunk with the same string minus the last character,
        # which is always '\r' in this case.
        chunks[-1] = chunks[-1][:-1]
    else:
        # Just remove the "\r\n" from the latest chunk
        chunks.append(buf[: rlen - 2])

    return buf[rlen:], b"".join(chunks)


def _readsegment(
    sock: socket.socket, buf: bytes, end_tokens: bytes
) -> Tuple[bytes, bytes]:
    """Read a segment from the socket.

    Read a segment from the socket, up to the first end_token sub-string/bytes,
    and return that segment.

    Args:
        sock: Socket object, should be connected.
        buf: bytes, zero or more bytes, returned from an earlier
            call to _readline, _readsegment or _readvalue (pass an empty
            byte-string on the first call).
        end_tokens: bytes, indicates the end of the segment, generally this is
            b"\\r\\n" for memcached.

    Returns:
      A tuple of (buf, line) where line is the full line read from the
      socket (minus the end_tokens bytes) and buf is any trailing
      characters read after the end_tokens was found (which may be an empty
      bytes object).

    """
    result = bytes()

    while True:

        tokens_pos = buf.find(end_tokens)
        if tokens_pos != -1:
            before, after = buf[:tokens_pos], buf[tokens_pos + len(end_tokens) :]
            result += before
            return after, result

        buf = _recv(sock, RECV_SIZE)
        if not buf:
            raise MemcacheUnexpectedCloseError()


def _recv(sock: socket.socket, size: int) -> bytes:
    """sock.recv() with retry on EINTR"""
    while True:
        try:
            return sock.recv(size)
        except OSError as e:
            if e.errno != errno.EINTR:
                raise



""" Module containing the RetryingClient wrapper class. """

from time import sleep


def _ensure_tuple_argument(argument_name, argument_value):
    """
    Helper function to ensure the given arguments are tuples of Exceptions (or
    subclasses), or can at least be converted to such.

    Args:
      argument_name: str, name of the argument we're checking, only used for
        raising meaningful exceptions.
      argument: any, the argument itself.

    Returns:
      tuple[Exception]: A tuple with the elements from the argument if they are
        valid.

    Exceptions:
      ValueError: If the argument was not None, tuple or Iterable.
      ValueError: If any of the elements of the argument is not a subclass of
        Exception.
    """

    # Ensure the argument is a tuple, set or list.
    if argument_value is None:
        return tuple()
    elif not isinstance(argument_value, (tuple, set, list)):
        raise ValueError("%s must be either a tuple, a set or a list." % argument_name)

    # Convert the argument before checking contents.
    argument_tuple = tuple(argument_value)

    # Check that all the elements are actually inherited from Exception.
    # (Catchable)
    if not all([issubclass(arg, Exception) for arg in argument_tuple]):
        raise ValueError(
            "%s is only allowed to contain elements that are subclasses of "
            "Exception." % argument_name
        )

    return argument_tuple


class RetryingClient(object):
    """
    Client that allows retrying calls for the other clients.
    """

    def __init__(
        self, client, attempts=2, retry_delay=0, retry_for=None, do_not_retry_for=None
    ):
        """
        Constructor for RetryingClient.

        Args:
          client: Client|PooledClient|HashClient, inner client to use for
            performing actual work.
          attempts: optional int, how many times to attempt an action before
            failing. Must be 1 or above. Defaults to 2.
          retry_delay: optional int|float, how many seconds to sleep between
            each attempt.
            Defaults to 0.

          retry_for: optional None|tuple|set|list, what exceptions to
            allow retries for. Will allow retries for all exceptions if None.
            Example:
                `(MemcacheClientError, MemcacheUnexpectedCloseError)`
            Accepts any class that is a subclass of Exception.
            Defaults to None.

          do_not_retry_for: optional None|tuple|set|list, what
            exceptions should be retried. Will not block retries for any
            Exception if None.
            Example:
                `(IOError, MemcacheIllegalInputError)`
            Accepts any class that is a subclass of Exception.
            Defaults to None.

        Exceptions:
          ValueError: If `attempts` is not 1 or above.
          ValueError: If `retry_for` or `do_not_retry_for` is not None, tuple or
            Iterable.
          ValueError: If any of the elements of `retry_for` or
            `do_not_retry_for` is not a subclass of Exception.
          ValueError: If there is any overlap between `retry_for` and
            `do_not_retry_for`.
        """

        if attempts < 1:
            raise ValueError(
                "`attempts` argument must be at least 1. "
                "Otherwise no attempts are made."
            )

        self._client = client
        self._attempts = attempts
        self._retry_delay = retry_delay
        self._retry_for = _ensure_tuple_argument("retry_for", retry_for)
        self._do_not_retry_for = _ensure_tuple_argument(
            "do_not_retry_for", do_not_retry_for
        )

        # Verify no overlap in the go/no-go exception collections.
        for exc_class in self._retry_for:
            if exc_class in self._do_not_retry_for:
                raise ValueError(
                    'Exception class "%s" was present in both `retry_for` '
                    "and `do_not_retry_for`. Any exception class is only "
                    "allowed in a single argument." % repr(exc_class)
                )

        # Take dir from the client to speed up future checks.
        self._client_dir = dir(self._client)

    def _retry(self, name, func, *args, **kwargs):
        """
        Workhorse function, handles retry logic.

        Args:
          name: str, Name of the function called.
          func: callable, the function to retry.
          *args: args, array arguments to pass to the function.
          **kwargs: kwargs, keyword arguments to pass to the function.
        """
        for attempt in range(self._attempts):
            try:
                result = func(*args, **kwargs)
                return result

            except Exception as exc:
                # Raise the exception to caller if either is met:
                # - We've used the last attempt.
                # - self._retry_for is set, and we do not match.
                # - self._do_not_retry_for is set, and we do match.
                # - name is not actually a member of the client class.
                if (
                    attempt >= self._attempts - 1
                    or (self._retry_for and not isinstance(exc, self._retry_for))
                    or (
                        self._do_not_retry_for
                        and isinstance(exc, self._do_not_retry_for)
                    )
                    or name not in self._client_dir
                ):
                    raise exc

                # Sleep and try again.
                sleep(self._retry_delay)

    # This is the real magic soup of the class, we catch anything that isn't
    # strictly defined for ourselves and pass it on to whatever client we've
    # been given.
    def __getattr__(self, name):

        return lambda *args, **kwargs: self._retry(
            name, self._client.__getattribute__(name), *args, **kwargs
        )

    # We implement these explicitly because they're "magic" functions and won't
    # get passed on by __getattr__.

    def __dir__(self):
        return self._client_dir

    # These magics are copied from the base client.
    def __setitem__(self, key, value):
        self.set(key, value, noreply=True)

    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError
        return value

    def __delitem__(self, key):
        self.delete(key, noreply=True)



"""
Useful testing utilities.

This module is considered public API.

"""

import time

import socket

from pymemcache.exceptions import MemcacheClientError, MemcacheIllegalInputError
from pymemcache.serde import LegacyWrappingSerde
from pymemcache.client.base import check_key_helper


class MockMemcacheClient:
    """
    A (partial) in-memory mock for Clients.

    """

    def __init__(
        self,
        server=None,
        serde=None,
        serializer=None,
        deserializer=None,
        connect_timeout=None,
        timeout=None,
        no_delay=False,
        ignore_exc=False,
        socket_module=None,
        default_noreply=True,
        allow_unicode_keys=False,
        encoding="ascii",
        tls_context=None,
    ):

        self._contents = {}

        self.serde = serde or LegacyWrappingSerde(serializer, deserializer)
        self.allow_unicode_keys = allow_unicode_keys

        # Unused, but present for interface compatibility
        self.server = server
        self.connect_timeout = connect_timeout
        self.timeout = timeout
        self.no_delay = no_delay
        self.ignore_exc = ignore_exc
        self.socket_module = socket
        self.sock = None
        self.encoding = encoding
        self.tls_context = tls_context

    def check_key(self, key):
        """Checks key and add key_prefix."""
        return check_key_helper(key, allow_unicode_keys=self.allow_unicode_keys)

    def clear(self):
        """Method used to clear/reset mock cache"""
        self._contents.clear()

    def get(self, key, default=None):
        key = self.check_key(key)

        if key not in self._contents:
            return default

        expire, value, flags = self._contents[key]
        if expire and expire < time.time():
            del self._contents[key]
            return default

        return self.serde.deserialize(key, value, flags)

    def get_many(self, keys):
        out = {}
        for key in keys:
            value = self.get(key)
            if value is not None:
                out[key] = value
        return out

    get_multi = get_many

    def set(self, key, value, expire=0, noreply=True, flags=None):
        key = self.check_key(key)
        if isinstance(value, str) and not isinstance(value, bytes):
            try:
                value = value.encode(self.encoding)
            except (UnicodeEncodeError, UnicodeDecodeError):
                raise MemcacheIllegalInputError

        value, flags = self.serde.serialize(key, value)

        if expire:
            expire += time.time()

        self._contents[key] = expire, value, flags
        return True

    def set_many(self, values, expire=0, noreply=True, flags=None):
        result = []
        for key, value in values.items():
            ret = self.set(key, value, expire, noreply, flags=flags)
            if not ret:
                result.append(key)
        return [] if noreply else result

    set_multi = set_many

    def incr(self, key, value, noreply=False):
        current = self.get(key)
        present = current is not None
        if present:
            self.set(key, current + value, noreply=noreply)
        return None if noreply or not present else current + value

    def decr(self, key, value, noreply=False):
        current = self.get(key)
        present = current is not None
        if present:
            self.set(key, current - value, noreply=noreply)
        return None if noreply or not present else current - value

    def add(self, key, value, expire=0, noreply=True, flags=None):
        current = self.get(key)
        present = current is not None
        if not present:
            self.set(key, value, expire, noreply, flags=flags)
        return noreply or not present

    def delete(self, key, noreply=True):
        key = self.check_key(key)
        current = self._contents.pop(key, None)
        present = current is not None
        return noreply or present

    def delete_many(self, keys, noreply=True):
        for key in keys:
            self.delete(key, noreply)
        return True

    def prepend(self, key, value, expire=0, noreply=True, flags=None):
        current = self.get(key)
        if current is not None:
            if isinstance(value, str) and not isinstance(value, bytes):
                try:
                    value = value.encode(self.encoding)
                except (UnicodeEncodeError, UnicodeDecodeError):
                    raise MemcacheIllegalInputError
            self.set(key, value + current, expire, noreply, flags=flags)
        return True

    def append(self, key, value, expire=0, noreply=True, flags=None):
        current = self.get(key)
        if current is not None:
            if isinstance(value, str) and not isinstance(value, bytes):
                try:
                    value = value.encode(self.encoding)
                except (UnicodeEncodeError, UnicodeDecodeError):
                    raise MemcacheIllegalInputError
            self.set(key, current + value, expire, noreply, flags=flags)
        return True

    delete_multi = delete_many

    def stats(self, *_args):
        # I make no claim that these values make any sense, but the format
        # of the output is the same as for pymemcache.client.Client.stats()
        return {
            "version": "MockMemcacheClient",
            "rusage_user": 1.0,
            "rusage_system": 1.0,
            "hash_is_expanding": False,
            "slab_reassign_running": False,
            "inter": "in-memory",
            "evictions": False,
            "growth_factor": 1.0,
            "stat_key_prefix": "",
            "umask": 0o644,
            "detail_enabled": False,
            "cas_enabled": False,
            "auth_enabled_sasl": False,
            "maxconns_fast": False,
            "slab_reassign": False,
            "slab_automove": False,
        }

    def replace(self, key, value, expire=0, noreply=True, flags=None):
        current = self.get(key)
        present = current is not None
        if present:
            self.set(key, value, expire, noreply, flags=flags)
        return noreply or present

    def cas(self, key, value, cas, expire=0, noreply=False, flags=None):
        raise MemcacheClientError("CAS is not enabled for this instance")

    def touch(self, key, expire=0, noreply=True):
        current = self.get(key)
        present = current is not None
        if present:
            self.set(key, current, expire, noreply=noreply)
        return True if noreply or present else False

    def cache_memlimit(self, memlimit):
        return True

    def version(self):
        return "MockMemcacheClient"

    def flush_all(self, delay=0, noreply=True):
        self.clear()

        return noreply or self._contents == {}

    def quit(self):
        pass

    def close(self):
        pass



from pymemcache.client.rendezvous import RendezvousHash
import pytest


@pytest.mark.unit()
def test_init_no_options():
    rendezvous = RendezvousHash()
    assert 0 == len(rendezvous.nodes)
    assert 1361238019 == rendezvous.hash_function("6666")


@pytest.mark.unit()
def test_init():
    nodes = ["0", "1", "2"]
    rendezvous = RendezvousHash(nodes=nodes)
    assert 3 == len(rendezvous.nodes)
    assert 1361238019 == rendezvous.hash_function("6666")


@pytest.mark.unit()
def test_seed():
    rendezvous = RendezvousHash(seed=10)
    assert 2981722772 == rendezvous.hash_function("6666")


@pytest.mark.unit()
def test_add_node():
    rendezvous = RendezvousHash()
    rendezvous.add_node("1")

    assert 1 == len(rendezvous.nodes)
    rendezvous.add_node("1")

    assert 1 == len(rendezvous.nodes)
    rendezvous.add_node("2")

    assert 2 == len(rendezvous.nodes)
    rendezvous.add_node("1")

    assert 2 == len(rendezvous.nodes)


@pytest.mark.unit()
def test_remove_node():
    nodes = ["0", "1", "2"]
    rendezvous = RendezvousHash(nodes=nodes)
    rendezvous.remove_node("2")

    assert 2 == len(rendezvous.nodes)

    with pytest.raises(ValueError):
        rendezvous.remove_node("2")

    assert 2 == len(rendezvous.nodes)

    rendezvous.remove_node("1")
    assert 1 == len(rendezvous.nodes)

    rendezvous.remove_node("0")
    assert 0 == len(rendezvous.nodes)


@pytest.mark.unit()
def test_get_node():
    nodes = ["0", "1", "2"]
    rendezvous = RendezvousHash(nodes=nodes)
    assert "0" == rendezvous.get_node("ok")
    assert "1" == rendezvous.get_node("mykey")
    assert "2" == rendezvous.get_node("wat")


@pytest.mark.unit()
def test_get_node_after_removal():
    nodes = ["0", "1", "2"]
    rendezvous = RendezvousHash(nodes=nodes)
    rendezvous.remove_node("1")

    assert "0" == rendezvous.get_node("ok")
    assert "0" == rendezvous.get_node("mykey")
    assert "2" == rendezvous.get_node("wat")


@pytest.mark.unit()
def test_get_node_after_addition():
    nodes = ["0", "1", "2"]
    rendezvous = RendezvousHash(nodes=nodes)
    assert "0" == rendezvous.get_node("ok")
    assert "1" == rendezvous.get_node("mykey")
    assert "2" == rendezvous.get_node("wat")
    assert "2" == rendezvous.get_node("lol")
    rendezvous.add_node("3")

    assert "0" == rendezvous.get_node("ok")
    assert "1" == rendezvous.get_node("mykey")
    assert "2" == rendezvous.get_node("wat")
    assert "3" == rendezvous.get_node("lol")


@pytest.mark.unit()
def test_grow():
    rendezvous = RendezvousHash()

    placements = {}

    for i in range(10):
        rendezvous.add_node(str(i))
        placements[str(i)] = []

    for i in range(1000):
        node = rendezvous.get_node(str(i))
        placements[node].append(i)

    new_placements = {}

    for i in range(20):
        rendezvous.add_node(str(i))
        new_placements[str(i)] = []

    for i in range(1000):
        node = rendezvous.get_node(str(i))
        new_placements[node].append(i)

    keys = [k for sublist in placements.values() for k in sublist]
    new_keys = [k for sublist in new_placements.values() for k in sublist]
    assert sorted(keys) == sorted(new_keys)

    added = 0
    removed = 0

    for node, assignments in new_placements.items():
        after = set(assignments)
        before = set(placements.get(node, []))
        removed += len(before.difference(after))
        added += len(after.difference(before))

    assert added == removed
    assert 1062 == (added + removed)


@pytest.mark.unit()
def test_shrink():
    rendezvous = RendezvousHash()

    placements = {}
    for i in range(10):
        rendezvous.add_node(str(i))
        placements[str(i)] = []

    for i in range(1000):
        node = rendezvous.get_node(str(i))
        placements[node].append(i)

    rendezvous.remove_node("9")
    new_placements = {}
    for i in range(9):
        new_placements[str(i)] = []

    for i in range(1000):
        node = rendezvous.get_node(str(i))
        new_placements[node].append(i)

    keys = [k for sublist in placements.values() for k in sublist]
    new_keys = [k for sublist in new_placements.values() for k in sublist]
    assert sorted(keys) == sorted(new_keys)

    added = 0
    removed = 0
    for node, assignments in placements.items():
        after = set(assignments)
        before = set(new_placements.get(node, []))
        removed += len(before.difference(after))
        added += len(after.difference(before))

    assert added == removed
    assert 202 == (added + removed)


def collide(key, seed):
    return 1337


@pytest.mark.unit()
def test_rendezvous_collision():
    nodes = ["c", "b", "a"]
    rendezvous = RendezvousHash(nodes, hash_function=collide)

    for i in range(1000):
        assert "c" == rendezvous.get_node(i)


@pytest.mark.unit()
def test_rendezvous_names():
    nodes = [1, 2, 3, "a", "b", "lol.wat.com"]
    rendezvous = RendezvousHash(nodes, hash_function=collide)

    for i in range(10):
        assert "lol.wat.com" == rendezvous.get_node(i)

    nodes = [1, "a", "0"]
    rendezvous = RendezvousHash(nodes, hash_function=collide)

    for i in range(10):
        assert "a" == rendezvous.get_node(i)



from pymemcache.client.murmur3 import murmur3_32


class RendezvousHash:
    """
    Implements the Highest Random Weight (HRW) hashing algorithm most
    commonly referred to as rendezvous hashing.

    Originally developed as part of python-clandestined.

    Copyright (c) 2014 Ernest W. Durbin III
    """

    def __init__(self, nodes=None, seed=0, hash_function=murmur3_32):
        """
        Constructor.
        """
        self.nodes = []
        self.seed = seed
        if nodes is not None:
            self.nodes = nodes
        self.hash_function = lambda x: hash_function(x, seed)

    def add_node(self, node):
        if node not in self.nodes:
            self.nodes.append(node)

    def remove_node(self, node):
        if node in self.nodes:
            self.nodes.remove(node)
        else:
            raise ValueError("No such node %s to remove" % (node))

    def get_node(self, key):
        high_score = -1
        winner = None

        for node in self.nodes:
            score = self.hash_function(f"{node}-{key}")

            if score > high_score:
                (high_score, winner) = (score, node)
            elif score == high_score:
                (high_score, winner) = (score, max(str(node), str(winner)))

        return winner



from unittest import TestCase

from pymemcache.serde import (
    CompressedSerde,
    pickle_serde,
    PickleSerde,
    FLAG_BYTES,
    FLAG_COMPRESSED,
    FLAG_PICKLE,
    FLAG_INTEGER,
    FLAG_TEXT,
)
import pytest
import pickle
import sys
import zlib


class CustomInt(int):
    """
    Custom integer type for testing.

    Entirely useless, but used to show that built in types get serialized and
    deserialized back as the same type of object.
    """

    pass


def check(serde, value, expected_flags):
    serialized, flags = serde.serialize(b"key", value)
    assert flags == expected_flags

    # pymemcache stores values as byte strings, so we immediately the value
    # if needed so deserialized works as it would with a real server
    if not isinstance(serialized, bytes):
        serialized = str(serialized).encode("ascii")

    deserialized = serde.deserialize(b"key", serialized, flags)
    assert deserialized == value


@pytest.mark.unit()
class TestSerde:
    serde = pickle_serde

    def test_bytes(self):
        check(self.serde, b"value", FLAG_BYTES)
        check(self.serde, b"\xc2\xa3 $ \xe2\x82\xac", FLAG_BYTES)  # £ $ €

    def test_unicode(self):
        check(self.serde, "value", FLAG_TEXT)
        check(self.serde, "£ $ €", FLAG_TEXT)

    def test_int(self):
        check(self.serde, 1, FLAG_INTEGER)

    def test_pickleable(self):
        check(self.serde, {"a": "dict"}, FLAG_PICKLE)

    def test_subtype(self):
        # Subclass of a native type will be restored as the same type
        check(self.serde, CustomInt(123123), FLAG_PICKLE)


@pytest.mark.unit()
class TestSerdePickleVersion0(TestCase):
    serde = PickleSerde(pickle_version=0)


@pytest.mark.unit()
class TestSerdePickleVersion1(TestCase):
    serde = PickleSerde(pickle_version=1)


@pytest.mark.unit()
class TestSerdePickleVersion2(TestCase):
    serde = PickleSerde(pickle_version=2)


@pytest.mark.unit()
class TestSerdePickleVersionHighest(TestCase):
    serde = PickleSerde(pickle_version=pickle.HIGHEST_PROTOCOL)


@pytest.mark.parametrize("serde", [pickle_serde, CompressedSerde()])
@pytest.mark.unit()
def test_compressed_simple(serde):
    # test_bytes
    check(serde, b"value", FLAG_BYTES)
    check(serde, b"\xc2\xa3 $ \xe2\x82\xac", FLAG_BYTES)  # £ $ €

    # test_unicode
    check(serde, "value", FLAG_TEXT)
    check(serde, "£ $ €", FLAG_TEXT)

    # test_int
    check(serde, 1, FLAG_INTEGER)

    # test_pickleable
    check(serde, {"a": "dict"}, FLAG_PICKLE)

    # test_subtype
    # Subclass of a native type will be restored as the same type
    check(serde, CustomInt(12312), FLAG_PICKLE)


@pytest.mark.parametrize(
    "serde",
    [
        CompressedSerde(min_compress_len=49),
        # Custom compression.  This could be something like lz4
        CompressedSerde(
            compress=lambda value: zlib.compress(value, 9),
            decompress=lambda value: zlib.decompress(value),
            min_compress_len=49,
        ),
    ],
)
@pytest.mark.unit()
def test_compressed_complex(serde):
    # test_bytes
    check(serde, b"value" * 10, FLAG_BYTES | FLAG_COMPRESSED)
    check(serde, b"\xc2\xa3 $ \xe2\x82\xac" * 10, FLAG_BYTES | FLAG_COMPRESSED)  # £ $ €

    # test_unicode
    check(serde, "value" * 10, FLAG_TEXT | FLAG_COMPRESSED)
    check(serde, "£ $ €" * 10, FLAG_TEXT | FLAG_COMPRESSED)

    # test_int, doesn't make sense to compress
    check(serde, sys.maxsize, FLAG_INTEGER)

    # test_pickleable
    check(
        serde,
        {
            "foo": "bar",
            "baz": "qux",
            "uno": "dos",
            "tres": "tres",
        },
        FLAG_PICKLE | FLAG_COMPRESSED,
    )

    # test_subtype
    # Subclass of a native type will be restored as the same type
    check(serde, CustomInt(sys.maxsize), FLAG_PICKLE | FLAG_COMPRESSED)



# Copyright 2012 Pinterest.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pickle
import zlib
from functools import partial
from io import BytesIO

FLAG_BYTES = 0
FLAG_PICKLE = 1 << 0
FLAG_INTEGER = 1 << 1
FLAG_LONG = 1 << 2
FLAG_COMPRESSED = 1 << 3
FLAG_TEXT = 1 << 4

# Pickle protocol version (highest available to runtime)
# Warning with `0`: If somewhere in your value lies a slotted object,
# ie defines `__slots__`, even if you do not include it in your pickleable
# state via `__getstate__`, python will complain with something like:
#   TypeError: a class that defines __slots__ without defining __getstate__
#   cannot be pickled
DEFAULT_PICKLE_VERSION = pickle.HIGHEST_PROTOCOL


def _python_memcache_serializer(key, value, pickle_version=None):
    flags = 0
    value_type = type(value)

    # Check against exact types so that subclasses of native types will be
    # restored as their native type
    if value_type is bytes:
        pass

    elif value_type is str:
        flags |= FLAG_TEXT
        value = value.encode("utf8")

    elif value_type is int:
        flags |= FLAG_INTEGER
        value = "%d" % value

    else:
        flags |= FLAG_PICKLE
        output = BytesIO()
        pickler = pickle.Pickler(output, pickle_version)
        pickler.dump(value)
        value = output.getvalue()

    return value, flags


def get_python_memcache_serializer(pickle_version: int = DEFAULT_PICKLE_VERSION):
    """Return a serializer using a specific pickle version"""
    return partial(_python_memcache_serializer, pickle_version=pickle_version)


python_memcache_serializer = get_python_memcache_serializer()


def python_memcache_deserializer(key, value, flags):
    if flags == 0:
        return value

    elif flags & FLAG_TEXT:
        return value.decode("utf8")

    elif flags & FLAG_INTEGER:
        return int(value)

    elif flags & FLAG_LONG:
        return int(value)

    elif flags & FLAG_PICKLE:
        try:
            buf = BytesIO(value)
            unpickler = pickle.Unpickler(buf)
            return unpickler.load()
        except Exception:
            logging.info("Pickle error", exc_info=True)
            return None

    return value


class PickleSerde:
    """
    An object which implements the serialization/deserialization protocol for
    :py:class:`pymemcache.client.base.Client` and its descendants using the
    :mod:`pickle` module.

    Serialization and deserialization are implemented as methods of this class.
    To implement a custom serialization/deserialization method for pymemcache,
    you should implement the same interface as the one provided by this object
    -- :py:meth:`pymemcache.serde.PickleSerde.serialize` and
    :py:meth:`pymemcache.serde.PickleSerde.deserialize`. Then,
    pass your custom object to the pymemcache client object in place of
    `PickleSerde`.

    For more details on the serialization protocol, see the class documentation
    for :py:class:`pymemcache.client.base.Client`
    """

    def __init__(self, pickle_version: int = DEFAULT_PICKLE_VERSION) -> None:
        self._serialize_func = get_python_memcache_serializer(pickle_version)

    def serialize(self, key, value):
        return self._serialize_func(key, value)

    def deserialize(self, key, value, flags):
        return python_memcache_deserializer(key, value, flags)


pickle_serde = PickleSerde()


class CompressedSerde:
    """
    An object which implements the serialization/deserialization protocol for
    :py:class:`pymemcache.client.base.Client` and its descendants with
    configurable compression.
    """

    def __init__(
        self,
        compress=zlib.compress,
        decompress=zlib.decompress,
        serde=pickle_serde,
        # Discovered via the `test_optimal_compression_length` test.
        min_compress_len=400,
    ):
        self._serde = serde
        self._compress = compress
        self._decompress = decompress
        self._min_compress_len = min_compress_len

    def serialize(self, key, value):
        value, flags = self._serde.serialize(key, value)

        if len(value) > self._min_compress_len > 0:
            old_value = value
            value = self._compress(value)
            # Don't use the compressed value if our end result is actually
            # larger uncompressed.
            if len(old_value) < len(value):
                value = old_value
            else:
                flags |= FLAG_COMPRESSED

        return value, flags

    def deserialize(self, key, value, flags):
        if flags & FLAG_COMPRESSED:
            value = self._decompress(value)

        value = self._serde.deserialize(key, value, flags)
        return value


compressed_serde = CompressedSerde()


class LegacyWrappingSerde:
    """
    This class defines how to wrap legacy de/serialization functions into a
    'serde' object which implements '.serialize' and '.deserialize' methods.
    It is used automatically by pymemcache.client.base.Client when the
    'serializer' or 'deserializer' arguments are given.

    The serializer_func and deserializer_func are expected to be None in the
    case that they are missing.
    """

    def __init__(self, serializer_func, deserializer_func) -> None:
        self.serialize = serializer_func or self._default_serialize
        self.deserialize = deserializer_func or self._default_deserialize

    def _default_serialize(self, key, value):
        return value, 0

    def _default_deserialize(self, key, value, flags):
        return value



from pymemcache.client.base import Client
from pymemcache.serde import (
    CompressedSerde,
    pickle_serde,
)

from faker import Faker

import pytest
import random
import string
import time
import zstd  # type: ignore
import zlib

fake = Faker(["it_IT", "en_US", "ja_JP"])


def get_random_string(length):
    letters = string.ascii_letters
    chars = string.punctuation
    digits = string.digits
    total = letters + chars + digits
    result_str = "".join(random.choice(total) for i in range(length))
    return result_str


class CustomObject:
    """
    Custom class for verifying serialization
    """

    def __init__(self):
        self.number = random.randint(0, 100)
        self.string = fake.text()
        self.object = fake.profile()


class CustomObjectValue:
    def __init__(self, value):
        self.value = value


def benchmark(count, func, *args, **kwargs):
    start = time.time()

    for _ in range(count):
        result = func(*args, **kwargs)

    duration = time.time() - start
    print(str(duration))

    return result


@pytest.fixture(scope="session")
def names():
    names = []
    for _ in range(15):
        names.append(fake.name())

    return names


@pytest.fixture(scope="session")
def paragraphs():
    paragraphs = []
    for _ in range(15):
        paragraphs.append(fake.text())

    return paragraphs


@pytest.fixture(scope="session")
def objects():
    objects = []
    for _ in range(15):
        objects.append(CustomObject())

    return objects


# Always run compression for the benchmarks
min_compress_len = 1

default_serde = CompressedSerde(min_compress_len=min_compress_len)

zlib_serde = CompressedSerde(
    compress=lambda value: zlib.compress(value, 9),
    decompress=lambda value: zlib.decompress(value),
    min_compress_len=min_compress_len,
)

zstd_serde = CompressedSerde(
    compress=lambda value: zstd.compress(value),
    decompress=lambda value: zstd.decompress(value),
    min_compress_len=min_compress_len,
)

serializers = [
    None,
    default_serde,
    zlib_serde,
    zstd_serde,
]
ids = ["none", "zlib ", "zlib9", "zstd "]


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_set_strings(count, host, port, serde, names):
    client = Client((host, port), serde=serde, encoding="utf-8")

    def test():
        for index, name in enumerate(names):
            key = f"name_{index}"
            client.set(key, name)

    benchmark(count, test)


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_get_strings(count, host, port, serde, names):
    client = Client((host, port), serde=serde, encoding="utf-8")
    for index, name in enumerate(names):
        key = f"name_{index}"
        client.set(key, name)

    def test():
        for index, _ in enumerate(names):
            key = f"name_{index}"
            client.get(key)

    benchmark(count, test)


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_set_large_strings(count, host, port, serde, paragraphs):
    client = Client((host, port), serde=serde, encoding="utf-8")

    def test():
        for index, p in enumerate(paragraphs):
            key = f"paragraph_{index}"
            client.set(key, p)

    benchmark(count, test)


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_get_large_strings(count, host, port, serde, paragraphs):
    client = Client((host, port), serde=serde, encoding="utf-8")
    for index, p in enumerate(paragraphs):
        key = f"paragraphs_{index}"
        client.set(key, p)

    def test():
        for index, _ in enumerate(paragraphs):
            key = f"paragraphs_{index}"
            client.get(key)

    benchmark(count, test)


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_set_objects(count, host, port, serde, objects):
    client = Client((host, port), serde=serde, encoding="utf-8")

    def test():
        for index, o in enumerate(objects):
            key = f"objects_{index}"
            client.set(key, o)

    benchmark(count, test)


@pytest.mark.benchmark()
@pytest.mark.parametrize("serde", serializers, ids=ids)
def test_bench_compress_get_objects(count, host, port, serde, objects):
    client = Client((host, port), serde=serde, encoding="utf-8")
    for index, o in enumerate(objects):
        key = f"objects_{index}"
        client.set(key, o)

    def test():
        for index, _ in enumerate(objects):
            key = f"objects_{index}"
            client.get(key)

    benchmark(count, test)


@pytest.mark.benchmark()
def test_optimal_compression_length():
    for length in range(5, 2000):
        input_data = get_random_string(length)
        start = len(input_data)

        for index, serializer in enumerate(serializers[1:]):
            name = ids[index + 1]
            value, _ = serializer.serialize("foo", input_data)
            end = len(value)
            print(f"serializer={name}\t start={start}\t end={end}")


@pytest.mark.benchmark()
def test_optimal_compression_length_objects():
    for length in range(5, 2000):
        input_data = get_random_string(length)
        obj = CustomObjectValue(input_data)
        start = len(pickle_serde.serialize("foo", obj)[0])

        for index, serializer in enumerate(serializers[1:]):
            name = ids[index + 1]
            value, _ = serializer.serialize("foo", obj)
            end = len(value)
            print(f"serializer={name}\t start={start}\t end={end}")



# Copyright 2012 Pinterest.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import defaultdict

import pytest
from pymemcache.client.base import Client
from pymemcache.exceptions import (
    MemcacheClientError,
    MemcacheIllegalInputError,
    MemcacheServerError,
)
from pymemcache.serde import PickleSerde, compressed_serde, pickle_serde


def get_set_helper(client, key, value, key2, value2):
    result = client.get(key)
    assert result is None

    client.set(key, value, noreply=False)
    result = client.get(key)
    assert result == value

    client.set(key2, value2, noreply=True)
    result = client.get(key2)
    assert result == value2

    result = client.get_many([key, key2])
    assert result == {key: value, key2: value2}

    result = client.get_many([])
    assert result == {}


@pytest.mark.integration()
@pytest.mark.parametrize(
    "serde",
    [
        pickle_serde,
        compressed_serde,
    ],
)
def test_get_set(client_class, host, port, serde, socket_module, key_prefix):
    client = client_class(
        (host, port), serde=serde, socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    key = b"key"
    value = b"value"
    key2 = b"key2"
    value2 = b"value2"
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "serde",
    [
        pickle_serde,
        compressed_serde,
    ],
)
def test_get_set_unicode_key(
    client_class, host, port, serde, socket_module, key_prefix
):
    client = client_class(
        (host, port),
        serde=serde,
        socket_module=socket_module,
        allow_unicode_keys=True,
        key_prefix=key_prefix,
    )
    client.flush_all()

    key = "こんにちは"
    value = b"hello"
    key2 = "my☃"
    value2 = b"value2"
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "serde",
    [
        pickle_serde,
        compressed_serde,
    ],
)
def test_add_replace(client_class, host, port, serde, socket_module, key_prefix):
    client = client_class(
        (host, port), serde=serde, socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    result = client.add(b"key", b"value", noreply=False)
    assert result is True
    result = client.get(b"key")
    assert result == b"value"

    result = client.add(b"key", b"value2", noreply=False)
    assert result is False
    result = client.get(b"key")
    assert result == b"value"

    result = client.replace(b"key1", b"value1", noreply=False)
    assert result is False
    result = client.get(b"key1")
    assert result is None

    result = client.replace(b"key", b"value2", noreply=False)
    assert result is True
    result = client.get(b"key")
    assert result == b"value2"


@pytest.mark.integration()
def test_append_prepend(client_class, host, port, socket_module, key_prefix):
    client = client_class(
        (host, port), socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    result = client.append(b"key", b"value", noreply=False)
    assert result is False
    result = client.get(b"key")
    assert result is None

    result = client.set(b"key", b"value", noreply=False)
    assert result is True
    result = client.append(b"key", b"after", noreply=False)
    assert result is True
    result = client.get(b"key")
    assert result == b"valueafter"

    result = client.prepend(b"key1", b"value", noreply=False)
    assert result is False
    result = client.get(b"key1")
    assert result is None

    result = client.prepend(b"key", b"before", noreply=False)
    assert result is True
    result = client.get(b"key")
    assert result == b"beforevalueafter"


@pytest.mark.integration()
def test_cas(client_class, host, port, socket_module, key_prefix):
    client = client_class(
        (host, port), socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()
    result = client.cas(b"key", b"value", b"1", noreply=False)
    assert result is None

    result = client.set(b"key", b"value", noreply=False)
    assert result is True

    # binary, string, and raw int all match -- should all be encoded as b'1'
    result = client.cas(b"key", b"value", b"1", noreply=False)
    assert result is False
    result = client.cas(b"key", b"value", "1", noreply=False)
    assert result is False
    result = client.cas(b"key", b"value", 1, noreply=False)
    assert result is False

    result, cas = client.gets(b"key")
    assert result == b"value"

    result = client.cas(b"key", b"value1", cas, noreply=False)
    assert result is True

    result = client.cas(b"key", b"value2", cas, noreply=False)
    assert result is False


@pytest.mark.integration()
def test_gets(client_class, host, port, socket_module, key_prefix):
    client = client_class(
        (host, port), socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    result = client.gets(b"key")
    assert result == (None, None)

    result = client.set(b"key", b"value", noreply=False)
    assert result is True
    result = client.gets(b"key")
    assert result[0] == b"value"


@pytest.mark.integration()
def test_delete(client_class, host, port, socket_module, key_prefix):
    client = client_class(
        (host, port), socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    result = client.delete(b"key", noreply=False)
    assert result is False

    result = client.get(b"key")
    assert result is None
    result = client.set(b"key", b"value", noreply=False)
    assert result is True
    result = client.delete(b"key", noreply=False)
    assert result is True
    result = client.get(b"key")
    assert result is None


@pytest.mark.integration()
def test_incr_decr(client_class, host, port, socket_module, key_prefix):
    client = Client((host, port), socket_module=socket_module, key_prefix=key_prefix)
    client.flush_all()

    result = client.incr(b"key", 1, noreply=False)
    assert result is None

    result = client.set(b"key", b"0", noreply=False)
    assert result is True
    result = client.incr(b"key", 1, noreply=False)
    assert result == 1

    def _bad_int():
        client.incr(b"key", b"foobar")

    with pytest.raises(MemcacheClientError):
        _bad_int()

    result = client.decr(b"key1", 1, noreply=False)
    assert result is None

    result = client.decr(b"key", 1, noreply=False)
    assert result == 0
    result = client.get(b"key")
    assert result == b"0"


@pytest.mark.integration()
def test_touch(client_class, host, port, socket_module, key_prefix):
    client = client_class(
        (host, port), socket_module=socket_module, key_prefix=key_prefix
    )
    client.flush_all()

    result = client.touch(b"key", noreply=False)
    assert result is False

    result = client.set(b"key", b"0", 1, noreply=False)
    assert result is True

    result = client.touch(b"key", noreply=False)
    assert result is True

    result = client.touch(b"key", 1, noreply=False)
    assert result is True


@pytest.mark.integration()
def test_misc(client_class, host, port, socket_module, key_prefix):
    client = Client((host, port), socket_module=socket_module, key_prefix=key_prefix)
    client.flush_all()

    # Ensure no exceptions are thrown
    client.stats("cachedump", "1", "1")

    success = client.cache_memlimit(50)
    assert success


@pytest.mark.integration()
def test_serialization_deserialization(host, port, socket_module):
    class JsonSerde:
        def serialize(self, key, value):
            return json.dumps(value).encode("ascii"), 1

        def deserialize(self, key, value, flags):
            if flags == 1:
                return json.loads(value.decode("ascii"))
            return value

    client = Client((host, port), serde=JsonSerde(), socket_module=socket_module)
    client.flush_all()

    value = {"a": "b", "c": ["d"]}
    client.set(b"key", value)
    result = client.get(b"key")
    assert result == value


def serde_serialization_helper(client_class, host, port, socket_module, serde):
    def check(value):
        client.set(b"key", value, noreply=False)
        result = client.get(b"key")
        assert result == value
        assert type(result) is type(value)

    client = client_class((host, port), serde=serde, socket_module=socket_module)
    client.flush_all()

    check(b"byte string")
    check("unicode string")
    check("olé")
    check("olé")
    check(1)
    check(123123123123123123123)
    check({"a": "pickle"})
    check(["one pickle", "two pickle"])
    testdict = defaultdict(int)
    testdict["one pickle"]
    testdict[b"two pickle"]
    check(testdict)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "serde",
    [
        pickle_serde,
        compressed_serde,
    ],
)
def test_serde_serialization(client_class, host, port, socket_module, serde):
    serde_serialization_helper(client_class, host, port, socket_module, serde)


@pytest.mark.integration()
def test_serde_serialization0(client_class, host, port, socket_module):
    serde_serialization_helper(
        client_class, host, port, socket_module, PickleSerde(pickle_version=0)
    )


@pytest.mark.integration()
def test_serde_serialization2(client_class, host, port, socket_module):
    serde_serialization_helper(
        client_class, host, port, socket_module, PickleSerde(pickle_version=2)
    )


@pytest.mark.integration()
def test_errors(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    def _key_with_ws():
        client.set(b"key with spaces", b"value", noreply=False)

    with pytest.raises(MemcacheIllegalInputError):
        _key_with_ws()

    def _key_with_illegal_carriage_return():
        client.set(b"\r\nflush_all", b"value", noreply=False)

    with pytest.raises(MemcacheIllegalInputError):
        _key_with_illegal_carriage_return()

    def _key_too_long():
        client.set(b"x" * 1024, b"value", noreply=False)

    with pytest.raises(MemcacheClientError):
        _key_too_long()

    def _unicode_key_in_set():
        client.set("\u0FFF", b"value", noreply=False)

    with pytest.raises(MemcacheClientError):
        _unicode_key_in_set()

    def _unicode_key_in_get():
        client.get("\u0FFF")

    with pytest.raises(MemcacheClientError):
        _unicode_key_in_get()

    def _unicode_value_in_set():
        client.set(b"key", "\u0FFF", noreply=False)

    with pytest.raises(MemcacheClientError):
        _unicode_value_in_set()


@pytest.mark.skip("https://github.com/pinterest/pymemcache/issues/39")
@pytest.mark.integration()
def test_tls(client_class, tls_host, tls_port, socket_module, tls_context):
    client = client_class(
        (tls_host, tls_port), socket_module=socket_module, tls_context=tls_context
    )
    client.flush_all()

    key = b"key"
    value = b"value"
    key2 = b"key2"
    value2 = b"value2"
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "serde,should_fail",
    [
        (pickle_serde, True),
        (compressed_serde, False),
    ],
)
def test_get_set_large(
    client_class,
    host,
    port,
    serde,
    socket_module,
    should_fail,
):
    client = client_class((host, port), serde=serde, socket_module=socket_module)
    client.flush_all()

    key = b"key"
    value = b"value" * 1024 * 1024
    key2 = b"key2"
    value2 = b"value2" * 1024 * 1024

    if should_fail:
        with pytest.raises(MemcacheServerError):
            get_set_helper(client, key, value, key2, value2)
    else:
        get_set_helper(client, key, value, key2, value2)
